import os
from dataclasses import asdict
from datetime import timedelta

from flytekit import Resources, current_context, dynamic, task, workflow
from flytekit.extras.accelerators import T4, GPUAccelerator
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile
from returns.result import Failure, Success

from pyrovelocity.interfaces import (
    DownloadDatasetInterface,
    PreprocessDataInterface,
    PyroVelocityTrainInterface,
)
from pyrovelocity.io.archive import create_tarball_from_filtered_dir
from pyrovelocity.io.gcs import upload_file_concurrently
from pyrovelocity.logging import configure_logging
from pyrovelocity.tasks.data import download_dataset
from pyrovelocity.tasks.postprocess import postprocess_dataset
from pyrovelocity.tasks.preprocess import preprocess_dataset
from pyrovelocity.tasks.summarize import summarize_dataset
from pyrovelocity.tasks.train import train_dataset
from pyrovelocity.utils import str_to_bool
from pyrovelocity.workflows.main_configuration import (
    PYROVELOCITY_SIMULATED_ONLY,
    PostprocessConfiguration,
    PostprocessOutputs,
    ResourcesJSON,
    SummarizeOutputs,
    TrainingOutputs,
    WorkflowConfiguration,
    default_resource_limits,
    default_resource_requests,
    default_training_resource_limits,
    default_training_resource_requests,
    larry_configuration,
    pancreas_configuration,
    pbmc68k_configuration,
    pons_configuration,
    simulated_configuration,
)

__all__ = [
    "download_data",
    "preprocess_data",
    "train_model",
    "postprocess_data",
    "summarize_data",
    "upload_summary",
    "module_workflow",
    "training_workflow",
]

logger = configure_logging(__name__)

CACHE_VERSION = "2024.6.27"
SUMMARIZE_CACHE_VERSION = "2024.6.27"
PYROVELOCITY_CACHE_FLAG = str_to_bool(
    os.getenv("PYROVELOCITY_CACHE_FLAG", "True")
)
L4 = GPUAccelerator("nvidia-l4")
ACCELERATOR_TYPE: GPUAccelerator = L4


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=CACHE_VERSION,
    retries=3,
    interruptible=True,
    timeout=timedelta(minutes=20),
    requests=Resources(cpu="2", mem="4Gi", ephemeral_storage="16Gi"),
    limits=Resources(cpu="8", mem="16Gi", ephemeral_storage="200Gi"),
)
def download_data(download_dataset_args: DownloadDatasetInterface) -> FlyteFile:
    """
    Download external data.
    """
    dataset_path = download_dataset(**asdict(download_dataset_args))
    print(f"\nFlyte download data path: {dataset_path}\n")
    return FlyteFile(path=dataset_path)


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=CACHE_VERSION,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=60),
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="32Gi"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi"),
)
def preprocess_data(
    data: FlyteFile, preprocess_data_args: PreprocessDataInterface
) -> FlyteFile:
    """
    Download external data.
    """
    data_path = data.download()
    print(f"\nFlyte preprocess input data path: {data_path}\n")
    preprocess_data_args.adata = str(data_path)
    _, processed_dataset_path = preprocess_dataset(
        **asdict(preprocess_data_args),
    )

    print(f"\nFlyte preprocess output data path: {processed_dataset_path}\n")
    return FlyteFile(path=processed_dataset_path)


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=CACHE_VERSION,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=120),
    container_image="{{.image.gpu.fqn}}:{{.image.gpu.version}}",
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="50Gi", gpu="1"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi", gpu="1"),
    accelerator=ACCELERATOR_TYPE,
)
def train_model(
    processed_data: FlyteFile,
    train_model_configuration: PyroVelocityTrainInterface,
) -> TrainingOutputs:
    """
    Train model.
    """
    processed_data_path = processed_data.download()
    print(f"\nFlyte train model input data path: {processed_data_path}\n")
    train_model_configuration.adata = str(processed_data_path)
    (
        data_model,
        data_model_path,
        trained_data_path,
        model_path,
        posterior_samples_path,
        metrics_path,
        run_info_path,
        loss_plot_path,
    ) = train_dataset(
        **asdict(train_model_configuration),
    )

    print(
        f"\nFlyte train model outputs:\n"
        f"\tdata model: {data_model}\n"
        f"\tdata model path: {data_model_path}\n"
        f"\ttrained data path: {trained_data_path}\n"
        f"\tmodel path: {model_path}\n"
        f"\tposterior samples path: {posterior_samples_path}\n"
        f"\tmetrics path: {metrics_path}\n"
        f"\trun info path: {run_info_path}\n"
        f"\tloss plot path: {loss_plot_path}\n\n"
    )

    return TrainingOutputs(
        data_model=data_model,
        data_model_path=data_model_path,
        trained_data_path=FlyteFile(path=trained_data_path),
        model_path=FlyteDirectory(path=model_path),
        posterior_samples_path=FlyteFile(path=posterior_samples_path),
        metrics_path=FlyteFile(path=metrics_path),
        run_info_path=FlyteFile(path=run_info_path),
        loss_plot_path=FlyteFile(path=loss_plot_path),
    )


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=CACHE_VERSION,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=120),
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="50Gi"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi"),
)
def postprocess_data(
    preprocess_data_args: PreprocessDataInterface,
    training_outputs: TrainingOutputs,
    postprocess_configuration: PostprocessConfiguration,
) -> PostprocessOutputs:
    trained_data_path = training_outputs.trained_data_path.download()
    model_path = training_outputs.model_path.download()
    posterior_samples_path = training_outputs.posterior_samples_path.download()
    metrics_path = training_outputs.metrics_path.download()

    pyrovelocity_data_path, postprocessed_data_path = postprocess_dataset(
        data_model=training_outputs.data_model,
        data_model_path=training_outputs.data_model_path,
        trained_data_path=trained_data_path,
        model_path=model_path,
        posterior_samples_path=posterior_samples_path,
        metrics_path=metrics_path,
        vector_field_basis=preprocess_data_args.vector_field_basis,
        number_posterior_samples=postprocess_configuration.number_posterior_samples,
    )

    print(
        f"\npostprocessed_data_path: {postprocessed_data_path}\n\n",
    )

    return PostprocessOutputs(
        pyrovelocity_data=FlyteFile(path=pyrovelocity_data_path),
        postprocessed_data=FlyteFile(path=postprocessed_data_path),
    )


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=SUMMARIZE_CACHE_VERSION,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=120),
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="50Gi"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi"),
)
def summarize_data(
    preprocess_data_args: PreprocessDataInterface,
    postprocessing_outputs: PostprocessOutputs,
    training_outputs: TrainingOutputs,
) -> SummarizeOutputs:
    model_path = training_outputs.model_path.download()
    pyrovelocity_data_path = postprocessing_outputs.pyrovelocity_data.download()
    postprocessed_data_path = (
        postprocessing_outputs.postprocessed_data.download()
    )

    print(
        f"\nmodel_path: {model_path}\n\n",
        f"pyrovelocity_data_path: {pyrovelocity_data_path}\n",
        f"postprocessed_data_path: {postprocessed_data_path}\n",
    )

    data_model_reports_path, dataframe_path = summarize_dataset(
        data_model=training_outputs.data_model,
        data_model_path=training_outputs.data_model_path,
        model_path=model_path,
        pyrovelocity_data_path=pyrovelocity_data_path,
        postprocessed_data_path=postprocessed_data_path,
        cell_state=preprocess_data_args.cell_state,
        vector_field_basis=preprocess_data_args.vector_field_basis,
    )
    print(
        f"\ndata_model_reports_path: {data_model_reports_path}\n",
        f"\ndataframe_path: {dataframe_path}\n\n",
    )
    return SummarizeOutputs(
        data_model_reports=FlyteDirectory(path=data_model_reports_path),
        dataframe=FlyteFile(path=dataframe_path),
    )


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=SUMMARIZE_CACHE_VERSION,
    retries=3,
    interruptible=True,
    timeout=timedelta(minutes=20),
    requests=Resources(cpu="2", mem="4Gi", ephemeral_storage="16Gi"),
    limits=Resources(cpu="8", mem="16Gi", ephemeral_storage="200Gi"),
)
def upload_summary(
    summarize_outputs: SummarizeOutputs, training_outputs: TrainingOutputs
) -> FlyteFile:
    data_model_reports = summarize_outputs.data_model_reports
    reports_path = data_model_reports.download()
    data_model = training_outputs.data_model

    ctx = current_context()
    execution_id = ctx.execution_id.name
    archive_name = f"archive_{data_model}_{execution_id}.tar.gz"

    print(f"\nCreating tarball\n{archive_name}\nfrom {reports_path}...\n\n")

    create_tarball_from_filtered_dir(
        reports_path,
        archive_name,
    )

    upload_result = upload_file_concurrently(
        f"pyrovelocity/reports/{execution_id}",
        archive_name,
        archive_name,
    )
    if isinstance(upload_result, Success):
        file_url = upload_result.unwrap()
        print(f"\nUpload successful.\nFile URL: {file_url}\n\n")
        return FlyteFile(path=archive_name)
    elif isinstance(upload_result, Failure):
        error = upload_result.failure()
        print(
            f"\nUpload of {archive_name} failed with exception: {error}\n"
            f"Returning empty file.\n\n"
        )
        return FlyteFile(path="")


@dynamic
def module_workflow(
    download_dataset_args: DownloadDatasetInterface = DownloadDatasetInterface(),
    preprocess_data_args: PreprocessDataInterface = PreprocessDataInterface(),
    train_model_configuration_1: PyroVelocityTrainInterface = PyroVelocityTrainInterface(),
    train_model_configuration_2: PyroVelocityTrainInterface = PyroVelocityTrainInterface(),
    postprocess_configuration: PostprocessConfiguration = PostprocessConfiguration(),
    train_model_resource_requests: ResourcesJSON = default_training_resource_requests,
    train_model_resource_limits: ResourcesJSON = default_training_resource_limits,
    postprocessing_resource_requests: ResourcesJSON = default_resource_requests,
    postprocessing_resource_limits: ResourcesJSON = default_resource_limits,
    summarizing_resource_requests: ResourcesJSON = default_resource_requests,
    summarizing_resource_limits: ResourcesJSON = default_resource_limits,
    upload_results: bool = True,
) -> list[SummarizeOutputs]:
    """
    Apply the primary workflow to a single dataset with multiple model
    configurations.

    Args:
        download_dataset_args (DownloadDatasetInterface, optional): Configuration for
            pyrovelocity.tasks.data.download_dataset. Defaults to DownloadDatasetInterface().
        preprocess_data_args (PreprocessDataInterface, optional): Configuration for
            pyrovelocity.tasks.preprocess.preprocess_dataset. Defaults to PreprocessDataInterface().
        train_model_configuration_1 (PyroVelocityTrainInterface, optional): Configuration
            for pyrovelocity.train.train_dataset. Defaults to PyroVelocityTrainInterface().
        train_model_configuration_2 (PyroVelocityTrainInterface, optional): Configuration
            for pyrovelocity.train.train_dataset. Defaults to PyroVelocityTrainInterface().
        postprocess_configuration (PostprocessConfiguration, optional): Configuration for
            pyrovelocity.postprocess.postprocess_dataset. Defaults to PostprocessConfiguration().
        train_model_resource_requests (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_training_resource_requests.
        train_model_resource_limits (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_training_resource_limits.
        postprocessing_resource_requests (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_resource_requests.
        postprocessing_resource_limits (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_resource_limits.
        summarizing_resource_requests (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_resource_requests.
        summarizing_resource_limits (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_resource_limits.

    Returns:
        list[FlyteFile]: Workflow outputs as flytekit.types.file.FlyteFile objects.
    """
    data = download_data(download_dataset_args=download_dataset_args)
    processed_data = preprocess_data(
        data=data, preprocess_data_args=preprocess_data_args
    )

    train_model_configurations = [
        train_model_configuration_1,
        train_model_configuration_2,
    ]

    model_outputs: list[TrainingOutputs] = []
    dataset_summaries: list[SummarizeOutputs] = []
    for train_model_configuration in train_model_configurations:
        model_output = train_model(
            processed_data=processed_data,
            train_model_configuration=train_model_configuration,
        ).with_overrides(
            requests=Resources(**asdict(train_model_resource_requests)),
            limits=Resources(**asdict(train_model_resource_limits)),
        )
        model_outputs.append(model_output)

        postprocessing_outputs = postprocess_data(
            preprocess_data_args=preprocess_data_args,
            training_outputs=model_output,
            postprocess_configuration=postprocess_configuration,
        ).with_overrides(
            requests=Resources(**asdict(postprocessing_resource_requests)),
            limits=Resources(**asdict(postprocessing_resource_limits)),
        )

        dataset_summary = summarize_data(
            preprocess_data_args=preprocess_data_args,
            postprocessing_outputs=postprocessing_outputs,
            training_outputs=model_output,
        ).with_overrides(
            requests=Resources(**asdict(summarizing_resource_requests)),
            limits=Resources(**asdict(summarizing_resource_limits)),
        )

        if upload_results:
            upload_summary(
                summarize_outputs=dataset_summary,
                training_outputs=model_output,
            )

        dataset_summaries.append(dataset_summary)

    return dataset_summaries


@dynamic
def training_workflow(
    simulated_configuration: WorkflowConfiguration = simulated_configuration,
    pancreas_configuration: WorkflowConfiguration = pancreas_configuration,
    pbmc68k_configuration: WorkflowConfiguration = pbmc68k_configuration,
    pons_configuration: WorkflowConfiguration = pons_configuration,
    larry_configuration: WorkflowConfiguration = larry_configuration,
) -> list[list[SummarizeOutputs]]:
    """
    Apply the primary workflow to a collection of configurations.
    Conditionally executes configurations based on the value of PYROVELOCITY_SIMULATED_ONLY.
    """
    results = []
    configurations = [
        (simulated_configuration, "simulated"),
    ]

    if not PYROVELOCITY_SIMULATED_ONLY:
        configurations += [
            (pancreas_configuration, "pancreas"),
            (pbmc68k_configuration, "pbmc68k"),
            (pons_configuration, "pons"),
            (larry_configuration, "larry"),
        ]

    for config, _ in configurations:
        result = module_workflow(
            download_dataset_args=config.download_dataset,
            preprocess_data_args=config.preprocess_data,
            train_model_configuration_1=config.training_configuration_1,
            train_model_configuration_2=config.training_configuration_2,
            postprocess_configuration=config.postprocess_configuration,
            train_model_resource_requests=config.training_resources_requests,
            train_model_resource_limits=config.training_resources_limits,
            postprocessing_resource_requests=config.postprocessing_resources_requests,
            postprocessing_resource_limits=config.postprocessing_resources_limits,
            summarizing_resource_requests=config.summarizing_resources_requests,
            summarizing_resource_limits=config.summarizing_resources_limits,
            upload_results=config.upload_results,
        )
        results.append(result)

    return results


if __name__ == "__main__":
    """
    The module workflow is applied to a single dataset together with a list of
    model configurations.

    There are three ways to execute the train_model() task, which impacts how
    subsequent tasks are executed. This can also be executed with the @workflow
    decorator for a single model configuration

    ```python
    train_model_configuration: PyroVelocityTrainInterface = PyroVelocityTrainInterface(),
    ) -> TrainingOutputs:
    ...
    model_outputs = train_model(
        data=processed_data,
        train_model_configuration=train_model_configuration,
    )
    ```

    or for multiple model configurations using a map task

    ```python
    import functools
    from flytekit.experimental import map_task

    partial_train_model = functools.partial(train_model, data=processed_data)
    model_outputs = map_task(partial_train_model)(
        train_model_args=train_model_configurations,
    )
    ```

    The dynamic workflow is preferred to support both mapping of tasks over
    multiple model configurations and overriding resources from configuration
    data.
    """
    print(f"Running module_workflow() { module_workflow() }")
    print(f"Running training_workflow() { training_workflow() }")



from typing import Optional
from typing import Tuple
from typing import Union

import pyro
import torch
from beartype import beartype
from jaxtyping import Float
from jaxtyping import jaxtyped
from pyro import poutine
from pyro.distributions import Bernoulli
from pyro.distributions import LogNormal
from pyro.distributions import Normal
from pyro.distributions import Poisson
from pyro.nn import PyroModule
from pyro.nn import PyroSample
from pyro.primitives import plate
from scvi.nn import Decoder
from torch.nn.functional import relu
from torch.nn.functional import softplus

from pyrovelocity.logging import configure_logging
from pyrovelocity.models._transcription_dynamics import mrna_dynamics


logger = configure_logging(__name__)

RNAInputType = Union[
    Float[torch.Tensor, ""],
    Float[torch.Tensor, "num_genes"],
    Float[torch.Tensor, "samples num_genes"],
]

RNAOutputType = Union[
    Float[torch.Tensor, "num_cells num_genes"],
    Float[torch.Tensor, "samples num_cells num_genes"],
]


class LogNormalModel(PyroModule):
    """
    A base class for pyrovelocity models.

    This class serves as the base class for constructing a pyrovelocity model.
    It provides basic methods for handling gene expression data, such as creating plates
    for cells and genes, encoding cell-specific features, and computing the likelihood
    of the observed data.

    Attributes:
        num_cells (int): The number of cells.
        num_genes (int): The number of genes.
        likelihood (str): The likelihood type for the model, defaults to "Poisson".
        plate_size (int): The size of the plate for the model, defaults to 2.

    Example:
        >>> from pyrovelocity.models._velocity_model import LogNormalModel
        >>> num_cells = 3
        >>> num_genes = 4
        >>> likelihood = "Poisson"
        >>> plate_size = 2
        >>> model = LogNormalModel(num_cells, num_genes, likelihood, plate_size)
        >>> logger.info(model)
        >>> assert model.num_cells == num_cells
        >>> assert model.num_genes == num_genes
        >>> assert model.likelihood == likelihood
        >>> assert model.plate_size == plate_size
    """

    @beartype
    def __init__(
        self,
        num_cells: int,
        num_genes: int,
        likelihood: str = "Poisson",
        plate_size: int = 2,
        correct_library_size: Union[bool, str] = True,
    ) -> None:
        """
        Initialize the LogNormalModel.

        TODO: make the likelihood parameter a sum type over supported distributions.

        Args:
            num_cells (int): number of cells.
            num_genes (int): number of genes.
            likelihood (str, optional): Type of likelihood to use. Defaults to "Poisson".
            plate_size (int, optional): Number of plates. Defaults to 2.
            correct_library_size (Union[bool, str], optional): Flag to control normalization by library size. Defaults to True.
        """
        logger.info("Initializing LogNormalModel")
        assert num_cells > 0 and num_genes > 0
        super().__init__()
        self.num_cells = num_cells
        self.num_genes = num_genes
        self.n_obs = None
        self.plate_size = plate_size
        self.correct_library_size = correct_library_size
        self.register_buffer("zero", torch.tensor(0.0))
        self.register_buffer("one", torch.tensor(1.0))
        self.likelihood = likelihood

    @beartype
    def __repr__(self) -> str:
        return (
            f"\nLogNormalModel(\n"
            f"\tnum_cells={self.num_cells}, \n"
            f"\tnum_genes={self.num_genes}, \n"
            f"\tlikelihood={self.likelihood}, \n"
            f"\tplate_size={self.plate_size}, \n"
            f"\tcorrect_library_size={self.correct_library_size}\n"
            f")\n"
        )

    @beartype
    def create_plates(
        self,
        u_obs: Optional[torch.Tensor] = None,
        s_obs: Optional[torch.Tensor] = None,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        u_log_library_loc: Optional[torch.Tensor] = None,
        s_log_library_loc: Optional[torch.Tensor] = None,
        u_log_library_scale: Optional[torch.Tensor] = None,
        s_log_library_scale: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
        cell_state: Optional[torch.Tensor] = None,
        time_info: Optional[torch.Tensor] = None,
    ) -> Tuple[plate, plate]:
        """
        Create cell and gene plates for the model. note that usage in
        pyro.infer.autoguide.AutoGuideList requires the set of arguments be
        equivalent to the model's forward method. See
        https://github.com/pyro-ppl/pyro/blob/670e9cb57003b52aea02deeb1316a42db3e1de1f/pyro/infer/autoguide/guides.py#L60-L63.

        Args:
            u_obs (Optional[torch.Tensor], optional): _description_. Defaults to None.
            s_obs (Optional[torch.Tensor], optional): _description_. Defaults to None.
            u_log_library (Optional[torch.Tensor], optional): _description_. Defaults to None.
            s_log_library (Optional[torch.Tensor], optional): _description_. Defaults to None.
            u_log_library_loc (Optional[torch.Tensor], optional): _description_. Defaults to None.
            s_log_library_loc (Optional[torch.Tensor], optional): _description_. Defaults to None.
            u_log_library_scale (Optional[torch.Tensor], optional): _description_. Defaults to None.
            s_log_library_scale (Optional[torch.Tensor], optional): _description_. Defaults to None.
            ind_x (Optional[torch.Tensor], optional): _description_. Defaults to None.
            cell_state (Optional[torch.Tensor], optional): _description_. Defaults to None.
            time_info (Optional[torch.Tensor], optional): _description_. Defaults to None.

        Returns:
            Tuple[plate, plate]: _description_
        """
        cell_plate = pyro.plate(
            "cells", self.num_cells, subsample=ind_x, dim=-2
        )
        gene_plate = pyro.plate("genes", self.num_genes, dim=-1)
        return cell_plate, gene_plate

    @PyroSample
    def alpha(self):
        return self._pyrosample_helper(1.0)

    @PyroSample
    def beta(self):
        return self._pyrosample_helper(0.25)

    @PyroSample
    def gamma(self):
        return self._pyrosample_helper(1.0)

    @PyroSample
    def u_scale(self):
        return self._pyrosample_helper(0.1)

    @PyroSample
    def s_scale(self):
        return self._pyrosample_helper(0.1)

    @PyroSample
    def u_inf(self):
        return self._pyrosample_helper(0.1)

    @PyroSample
    def s_inf(self):
        return self._pyrosample_helper(0.1)

    @PyroSample
    def dt_switching(self):
        return self._pyrosample_helper(1.0)

    @PyroSample
    def gene_offset(self):
        return Normal(self.zero, self.one)

    @PyroSample
    def t_scale(self):
        return Normal(self.zero, self.one * 0.1)

    @PyroSample
    def latent_time(self):
        if self.shared_time & self.plate_size == 2:
            return Normal(self.zero, self.one * 0.1).mask(self.include_prior)
        else:
            return (
                LogNormal(self.zero, self.one)
                .expand((self.num_genes,))
                .to_event(1)
            )

    @PyroSample
    def cell_time(self):
        if self.plate_size == 2 and self.shared_time:
            return Normal(self.zero, self.one)
        else:
            return LogNormal(self.zero, self.one)

    def _pyrosample_helper(self, scale: float):
        if self.plate_size == 2:
            return LogNormal(self.zero, self.one * scale)
        return (
            LogNormal(self.zero, self.one * 0.1)
            .expand((self.num_genes,))
            .to_event(1)
            .mask(False)
        )

    @beartype
    def get_likelihood(
        self,
        ut: torch.Tensor,
        st: torch.Tensor,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        u_scale: Optional[torch.Tensor] = None,
        s_scale: Optional[torch.Tensor] = None,
        u_read_depth: Optional[torch.Tensor] = None,
        s_read_depth: Optional[torch.Tensor] = None,
        u_cell_size_coef: None = None,
        ut_coef: None = None,
        s_cell_size_coef: None = None,
        st_coef: None = None,
    ) -> Tuple[Poisson, Poisson]:
        """
        Compute the likelihood of the given count data.

        Args:
            ut (torch.Tensor): Tensor representing unspliced transcripts.
            st (torch.Tensor): Tensor representing spliced transcripts.
            u_log_library (Optional[torch.Tensor], optional): Log library tensor for unspliced transcripts. Defaults to None.
            s_log_library (Optional[torch.Tensor], optional): Log library tensor for spliced transcripts. Defaults to None.
            u_scale (Optional[torch.Tensor], optional): Scale tensor for unspliced transcripts. Defaults to None.
            s_scale (Optional[torch.Tensor], optional): Scale tensor for spliced transcripts. Defaults to None.
            u_read_depth (Optional[torch.Tensor], optional): Read depth tensor for unspliced transcripts. Defaults to None.
            s_read_depth (Optional[torch.Tensor], optional): Read depth tensor for spliced transcripts. Defaults to None.
            u_cell_size_coef (Optional[Any], optional): Cell size coefficient for unspliced transcripts. Defaults to None.
            ut_coef (Optional[Any], optional): Coefficient for unspliced transcripts. Defaults to None.
            s_cell_size_coef (Optional[Any], optional): Cell size coefficient for spliced transcripts. Defaults to None.
            st_coef (Optional[Any], optional): Coefficient for spliced transcripts. Defaults to None.

        Returns:
            Tuple[Poisson, Poisson]: A tuple of Poisson distributions for unspliced and spliced transcripts, respectively.

        Example:
            >>> import torch
            >>> from pyrovelocity.models._velocity_model import LogNormalModel
            >>> num_cells = 10
            >>> num_genes = 20
            >>> likelihood = "Poisson"
            >>> plate_size = 2
            >>> model = LogNormalModel(num_cells, num_genes, likelihood, plate_size)
            >>> logger.info(model)
            >>> ut = torch.rand(num_cells, num_genes)
            >>> st = torch.rand(num_cells, num_genes)
            >>> u_read_depth = torch.rand(num_cells, 1)
            >>> s_read_depth = torch.rand(num_cells, 1)
            >>> u_dist, s_dist = model.get_likelihood(ut, st, u_read_depth=u_read_depth, s_read_depth=s_read_depth)
            >>> logger.info(f"u_dist: {u_dist}")
            >>> logger.info(f"s_dist: {s_dist}")
            >>> assert isinstance(u_dist, torch.distributions.Poisson)
            >>> assert isinstance(s_dist, torch.distributions.Poisson)
        """
        if self.likelihood != "Poisson":
            likelihood_not_implemented_msg = (
                "In the future, the likelihood will be referred to via a "
                "member of a sum type over supported distributions"
            )
            raise NotImplementedError(likelihood_not_implemented_msg)

        if self.correct_library_size:
            ut = relu(ut) + self.one * 1e-6
            st = relu(st) + self.one * 1e-6
            ut = pyro.deterministic("ut", ut, event_dim=0)
            st = pyro.deterministic("st", st, event_dim=0)
            ut = ut / torch.sum(ut, dim=-1, keepdim=True)
            st = st / torch.sum(st, dim=-1, keepdim=True)
            ut = pyro.deterministic("ut_norm", ut, event_dim=0)
            st = pyro.deterministic("st_norm", st, event_dim=0)
            ut = (ut + self.one * 1e-6) * u_read_depth
            st = (st + self.one * 1e-6) * s_read_depth
        else:
            ut = relu(ut)
            st = relu(st)
            ut = pyro.deterministic("ut", ut, event_dim=0)
            st = pyro.deterministic("st", st, event_dim=0)
            ut = ut + self.one * 1e-6
            st = st + self.one * 1e-6

        u_dist = Poisson(ut)
        s_dist = Poisson(st)
        return u_dist, s_dist


class VelocityModelAuto(LogNormalModel):
    """Automatically configured velocity model.

    Args:
        num_cells (int): _description_
        num_genes (int): _description_
        likelihood (str, optional): _description_. Defaults to "Poisson".
        shared_time (bool, optional): _description_. Defaults to True.
        t_scale_on (bool, optional): _description_. Defaults to False.
        plate_size (int, optional): _description_. Defaults to 2.
        latent_factor (str, optional): _description_. Defaults to "none".
        latent_factor_size (int, optional): _description_. Defaults to 30.
        latent_factor_operation (str, optional): _description_. Defaults to "selection".
        include_prior (bool, optional): _description_. Defaults to False.
        num_aux_cells (int, optional): _description_. Defaults to 100.
        only_cell_times (bool, optional): _description_. Defaults to False.
        decoder_on (bool, optional): _description_. Defaults to False.
        add_offset (bool, optional): _description_. Defaults to False.
        correct_library_size (Union[bool, str], optional): _description_. Defaults to True.
        guide_type (str, optional): _description_. Defaults to "velocity".
        cell_specific_kinetics (Optional[str], optional): _description_. Defaults to None.
        kinetics_num (Optional[int], optional): _description_. Defaults to None.

    Examples:
        >>> import torch
        >>> from pyrovelocity.models._velocity_model import VelocityModelAuto
        >>> model = VelocityModelAuto(
        ...             3,
        ...             4,
        ...             "Poisson",
        ...             True,
        ...             False,
        ...             2,
        ...             "none",
        ...             latent_factor_operation="selection",
        ...             latent_factor_size=10,
        ...             include_prior=False,
        ...             num_aux_cells=0,
        ...             only_cell_times=True,
        ...             decoder_on=False,
        ...             add_offset=False,
        ...             correct_library_size=True,
        ...             guide_type="auto_t0_constraint",
        ...             cell_specific_kinetics=None,
        ...             **{}
        ...         )
        >>> logger.info(model)
    """

    @beartype
    def __init__(
        self,
        num_cells: int,
        num_genes: int,
        likelihood: str = "Poisson",
        shared_time: bool = True,
        t_scale_on: bool = False,
        plate_size: int = 2,
        latent_factor: str = "none",
        latent_factor_size: int = 30,
        latent_factor_operation: str = "selection",
        include_prior: bool = False,
        num_aux_cells: int = 100,
        only_cell_times: bool = False,
        decoder_on: bool = False,
        add_offset: bool = False,
        correct_library_size: Union[bool, str] = True,
        guide_type: str = "velocity",
        cell_specific_kinetics: Optional[str] = None,
        kinetics_num: Optional[int] = None,
        **initial_values,
    ) -> None:
        assert num_cells > 0 and num_genes > 0
        super().__init__(num_cells, num_genes, likelihood, plate_size)
        self.num_aux_cells = num_aux_cells
        self.only_cell_times = only_cell_times
        self.guide_type = guide_type
        self.cell_specific_kinetics = cell_specific_kinetics
        self.k = kinetics_num

        self.mask = initial_values.get(
            "mask", torch.ones(self.num_cells, self.num_genes).bool()
        )
        for key in initial_values:
            self.register_buffer(f"{key}_init", initial_values[key])

        self.shared_time = shared_time
        self.t_scale_on = t_scale_on
        self.add_offset = add_offset
        self.plate_size = plate_size

        self.latent_factor = latent_factor
        self.latent_factor_size = latent_factor_size
        self.latent_factor_operation = latent_factor_operation
        self.include_prior = include_prior
        self.decoder_on = decoder_on
        self.correct_library_size = correct_library_size
        if self.decoder_on:
            self.decoder = Decoder(1, self.num_genes, n_layers=2)

        self.enumeration = "parallel"
        # self.set_enumeration_strategy()

    def sample_cell_gene_state(self, t, switching):
        return (
            pyro.sample(
                "cell_gene_state",
                Bernoulli(logits=t - switching),
                infer={"enumerate": self.enumeration},
            )
            == self.zero
        )

    @beartype
    def __repr__(self) -> str:
        return (
            f"\nVelocityModelAuto(\n"
            f"\tnum_cells={self.num_cells}, \n"
            f"\tnum_genes={self.num_genes}, \n"
            f'\tlikelihood="{self.likelihood}", \n'
            f"\tshared_time={self.shared_time}, \n"
            f"\tt_scale_on={self.t_scale_on}, \n"
            f"\tplate_size={self.plate_size}, \n"
            f'\tlatent_factor="{self.latent_factor}", \n'
            f"\tlatent_factor_size={self.latent_factor_size}, \n"
            f'\tlatent_factor_operation="{self.latent_factor_operation}", \n'
            f"\tinclude_prior={self.include_prior}, \n"
            f"\tnum_aux_cells={self.num_aux_cells}, \n"
            f"\tonly_cell_times={self.only_cell_times}, \n"
            f"\tdecoder_on={self.decoder_on}, \n"
            f"\tadd_offset={self.add_offset}, \n"
            f"\tcorrect_library_size={self.correct_library_size}, \n"
            f'\tguide_type="{self.guide_type}", \n'
            f"\tcell_specific_kinetics={self.cell_specific_kinetics}, \n"
            f"\tkinetics_num={self.k}\n"
            f")\n"
        )

    @jaxtyped(typechecker=beartype)
    def get_rna(
        self,
        u_scale: RNAInputType,
        s_scale: RNAInputType,
        alpha: RNAInputType,
        beta: RNAInputType,
        gamma: RNAInputType,
        t: Float[torch.Tensor, "num_cells time"],
        u0: RNAInputType,
        s0: RNAInputType,
        t0: RNAInputType,
        switching: Optional[RNAInputType] = None,
        u_inf: Optional[RNAInputType] = None,
        s_inf: Optional[RNAInputType] = None,
    ) -> Tuple[RNAOutputType, RNAOutputType]:
        """
        Computes the unspliced (u) and spliced (s) RNA expression levels given
        the model parameters.

        Args:
            u_scale (torch.Tensor): Scaling factor for unspliced expression.
            s_scale (torch.Tensor): Scaling factor for spliced expression.
            alpha (torch.Tensor): Transcription rate.
            beta (torch.Tensor): Splicing rate.
            gamma (torch.Tensor): Degradation rate.
            t (torch.Tensor): Cell time.
            u0 (torch.Tensor): Unspliced RNA initial expression.
            s0 (torch.Tensor): Spliced RNA initial expression.
            t0 (torch.Tensor): Initial cell time.
            switching (Optional[torch.Tensor], optional): Switching time. Default is None.
            u_inf (Optional[torch.Tensor], optional): Unspliced RNA expression at switching time. Default is None.
            s_inf (Optional[torch.Tensor], optional): Spliced RNA expression at switching time. Default is None.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: The unspliced (u) and spliced (s) RNA expression levels.


        Examples:
            >>> import torch
            >>> from pyrovelocity.models._velocity_model import VelocityModelAuto
            >>> model = VelocityModelAuto(
            ...             3,
            ...             4,
            ...             "Poisson",
            ...             True,
            ...             False,
            ...             2,
            ...             "none",
            ...             latent_factor_operation="selection",
            ...             latent_factor_size=10,
            ...             include_prior=False,
            ...             num_aux_cells=0,
            ...             only_cell_times=True,
            ...             decoder_on=False,
            ...             add_offset=False,
            ...             correct_library_size=True,
            ...             guide_type="auto_t0_constraint",
            ...             cell_specific_kinetics=None,
            ...             **{}
            ...         )
            >>> u, s = model.get_rna(
            ...            u_scale=torch.tensor([0.9793, 1.0567, 0.8610, 0.9304], device="cpu"),
            ...            s_scale=torch.tensor(1.),
            ...            alpha=torch.tensor([0.4869, 1.5997, 1.3962, 0.5038], device="cpu"),
            ...            beta=torch.tensor([0.5403, 1.1192, 0.9912, 1.1783], device="cpu"),
            ...            gamma=torch.tensor([1.9612, 0.5533, 2.1050, 4.9345], device="cpu"),
            ...            t=torch.tensor([[0.4230], [0.5119], [0.2689]], device="cpu"),
            ...            u0=torch.tensor(0.),
            ...            s0=torch.tensor(0.),
            ...            t0=torch.tensor([-0.4867, 0.5581, -0.6957, 0.6028], device="cpu"),
            ...            switching=torch.tensor([1.1886, 1.1227, 0.6789, 4.1003], device="cpu"),
            ...            u_inf=torch.tensor([0.5367, 0.6695, 1.0479, 0.4206], device="cpu"),
            ...            s_inf=torch.tensor([0.1132, 0.2100, 0.3750, 0.0999], device="cpu"),
            >>>        )
            >>> logger.info(f"u: {u}")
            >>> logger.info(f"s: {s}")
        """
        state = self.sample_cell_gene_state(t, switching)
        alpha_off = self.zero
        u0_vec = torch.where(state, u0, u_inf)
        s0_vec = torch.where(state, s0, s_inf)
        alpha_vec = torch.where(state, alpha, alpha_off)
        tau = softplus(torch.where(state, t - t0, t - switching))

        ut, st = mrna_dynamics(tau, u0_vec, s0_vec, alpha_vec, beta, gamma)
        ut = ut * u_scale / s_scale
        return ut, st

    @beartype
    def forward(
        self,
        u_obs: torch.Tensor,
        s_obs: torch.Tensor,
        u_log_library: Optional[torch.Tensor] = None,
        s_log_library: Optional[torch.Tensor] = None,
        u_log_library_loc: Optional[torch.Tensor] = None,
        s_log_library_loc: Optional[torch.Tensor] = None,
        u_log_library_scale: Optional[torch.Tensor] = None,
        s_log_library_scale: Optional[torch.Tensor] = None,
        ind_x: Optional[torch.Tensor] = None,
        cell_state: Optional[torch.Tensor] = None,
        time_info: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Defines the forward model, which computes the unspliced (u) and spliced
        (s) RNA expression levels given the observations and model parameters.

        Args:
            u_obs (Optional[torch.Tensor], optional): Observed unspliced RNA expression. Default is None.
            s_obs (Optional[torch.Tensor], optional): Observed spliced RNA expression. Default is None.
            u_log_library (Optional[torch.Tensor], optional): Log-transformed library size for unspliced RNA. Default is None.
            s_log_library (Optional[torch.Tensor], optional): Log-transformed library size for spliced RNA. Default is None.
            u_log_library_loc (Optional[torch.Tensor], optional): Mean of log-transformed library size for unspliced RNA. Default is None.
            s_log_library_loc (Optional[torch.Tensor], optional): Mean of log-transformed library size for spliced RNA. Default is None.
            u_log_library_scale (Optional[torch.Tensor], optional): Scale of log-transformed library size for unspliced RNA. Default is None.
            s_log_library_scale (Optional[torch.Tensor], optional): Scale of log-transformed library size for spliced RNA. Default is None.
            ind_x (Optional[torch.Tensor], optional): Indices for the cells. Default is None.
            cell_state (Optional[torch.Tensor], optional): Cell state information. Default is None.
            time_info (Optional[torch.Tensor], optional): Time information for the cells. Default is None.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: The unspliced (u) and spliced (s) RNA expression levels.

        Examples:
            >>> import torch
            >>> from pyrovelocity.models._velocity_model import VelocityModelAuto
            >>> u_obs=torch.tensor(
            ...     [[33.,  1.,  7.,  1.],
            ...     [12., 30., 11.,  3.],
            ...     [ 1.,  1.,  8.,  5.]],
            ...     device="cpu",
            >>> )
            >>> s_obs=torch.tensor(
            ...     [[32.0, 0.0, 6.0, 0.0],
            ...     [11.0, 29.0, 10.0, 2.0],
            ...     [0.0, 0.0, 7.0, 4.0]],
            ...     device="cpu",
            >>> )
            >>> u_log_library=torch.tensor([[3.7377], [4.0254], [2.7081]], device="cpu")
            >>> s_log_library=torch.tensor([[3.6376], [3.9512], [2.3979]], device="cpu")
            >>> u_log_library_loc=torch.tensor([[3.4904], [3.4904], [3.4904]], device="cpu")
            >>> s_log_library_loc=torch.tensor([[3.3289], [3.3289], [3.3289]], device="cpu")
            >>> u_log_library_scale=torch.tensor([[0.6926], [0.6926], [0.6926]], device="cpu")
            >>> s_log_library_scale=torch.tensor([[0.8214], [0.8214], [0.8214]], device="cpu")
            >>> ind_x=torch.tensor([2, 0, 1], device="cpu")
            >>> model = VelocityModelAuto(3,4)
            >>> u, s = model.forward(
            >>>            u_obs,
            >>>            s_obs,
            >>>            u_log_library,
            >>>            s_log_library,
            >>>            u_log_library_loc,
            >>>            s_log_library_loc,
            >>>            u_log_library_scale,
            >>>            s_log_library_scale,
            >>>            ind_x,
            >>>        )
            >>> u, s
            (tensor([[33.,  1.,  7.,  1.],
                    [12., 30., 11.,  3.],
                    [ 1.,  1.,  8.,  5.]]),
            tensor([[32.,  0.,  6.,  0.],
                    [11., 29., 10.,  2.],
                    [ 0.,  0.,  7.,  4.]]))
        """
        cell_plate, gene_plate = self.create_plates(
            u_obs,
            s_obs,
            u_log_library,
            s_log_library,
            u_log_library_loc,
            s_log_library_loc,
            u_log_library_scale,
            s_log_library_scale,
            ind_x,
            cell_state,
            time_info,
        )

        with gene_plate, poutine.mask(mask=self.include_prior):
            alpha = self.alpha
            gamma = self.gamma
            beta = self.beta

            if self.add_offset:
                u0 = pyro.sample("u_offset", LogNormal(self.zero, self.one))
                s0 = pyro.sample("s_offset", LogNormal(self.zero, self.one))
            else:
                s0 = u0 = self.zero

            t0 = pyro.sample("t0", Normal(self.zero, self.one))

            u_scale = self.u_scale
            s_scale = self.one

            dt_switching = self.dt_switching
            u_inf, s_inf = mrna_dynamics(
                dt_switching, u0, s0, alpha, beta, gamma
            )
            u_inf = pyro.deterministic("u_inf", u_inf, event_dim=0)
            s_inf = pyro.deterministic("s_inf", s_inf, event_dim=0)
            switching = pyro.deterministic(
                "switching", dt_switching + t0, event_dim=0
            )

        with cell_plate:
            t = pyro.sample(
                "cell_time",
                LogNormal(self.zero, self.one).mask(self.include_prior),
            )

        with cell_plate:
            u_cell_size_coef = ut_coef = s_cell_size_coef = st_coef = None
            u_read_depth = pyro.sample(
                "u_read_depth", LogNormal(u_log_library, u_log_library_scale)
            )
            s_read_depth = pyro.sample(
                "s_read_depth", LogNormal(s_log_library, s_log_library_scale)
            )
            with gene_plate:
                ut, st = self.get_rna(
                    u_scale,
                    s_scale,
                    alpha,
                    beta,
                    gamma,
                    t,
                    u0,
                    s0,
                    t0,
                    switching,
                    u_inf,
                    s_inf,
                )
                u_dist, s_dist = self.get_likelihood(
                    ut,
                    st,
                    u_log_library,
                    s_log_library,
                    u_scale,
                    s_scale,
                    u_read_depth=u_read_depth,
                    s_read_depth=s_read_depth,
                    u_cell_size_coef=u_cell_size_coef,
                    ut_coef=ut_coef,
                    s_cell_size_coef=s_cell_size_coef,
                    st_coef=st_coef,
                )
                u = pyro.sample("u", u_dist, obs=u_obs)
                s = pyro.sample("s", s_dist, obs=s_obs)
        return u, s



import os
import pickle
import sys
from statistics import harmonic_mean
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Union

import mlflow
import numpy as np
import pyro
import torch
from anndata import AnnData
from beartype import beartype
from numpy import ndarray
from scvi.data import AnnDataManager
from scvi.data._constants import _SETUP_ARGS_KEY
from scvi.data._constants import _SETUP_METHOD_NAME
from scvi.data.fields import LayerField
from scvi.data.fields import NumericalObsField
from scvi.model._utils import parse_device_args
from scvi.model.base import BaseModelClass
from scvi.model.base._utils import _initialize_model
from scvi.model.base._utils import _load_saved_files
from scvi.model.base._utils import _validate_var_names
from scvi.module.base import PyroBaseModuleClass

from pyrovelocity.analysis.analyze import compute_mean_vector_field
from pyrovelocity.analysis.analyze import compute_volcano_data
from pyrovelocity.analysis.analyze import vector_field_uncertainty
from pyrovelocity.logging import configure_logging
from pyrovelocity.models._trainer import VelocityTrainingMixin
from pyrovelocity.models._velocity_module import VelocityModule


__all__ = ["PyroVelocity"]


logger = configure_logging(__name__)


class PyroVelocity(VelocityTrainingMixin, BaseModelClass):
    """
    PyroVelocity is a class for constructing and training a Pyro model for
    probabilistic RNA velocity estimation. This model leverages the
    probabilistic programming language Pyro to estimate the parameters of models
    for the dynamics of RNA transcription, splicing, and degradation, providing
    the opportunity for insight into cellular states and associated state
    transitions. It makes use of AnnData, scvi-tools, and other scverse
    ecosystem libraries.

    Public methods include training the model with various configurations,
    generating posterior samples for further analysis, and saving/loading the
    model for reproducibility and further analysis.

    Attributes:
        use_gpu (str): Whether and which GPU to use.
        cell_specific_kinetics (Optional[str]): Type of cell-specific kinetics.
        k (Optional[int]): Number of kinetics.
        layers (List[str]): List of layers in the dataset.
        input_type (str): Type of input data.
        module (VelocityModule):
            The Pyro module used for the velocity estimation model.
        num_cells (int): Number of cells in the dataset.
        num_samples (int): Number of posterior samples to generate.
        _model_summary_string (str): Summary string for the model.
        init_params_ (Dict[str, Any]): Initial parameters for the model.

    For usage examples, including training the model and generating posterior
    samples, refer to the individual method docstrings.
    """

    """
    The `Methods` section is not supported by all documentation generators but
    is provided detached from the class docstring for reference. Please
    see the docstrings for each method for more details. This list may ignore
    unused or private methods.

    Methods:
        train:
            Trains the PyroVelocity model using the provided data and configuration.
        setup_anndata: 
            Set up AnnData object for compatibility with the scvi-tools
            model training interface.
        generate_posterior_samples:
            Generates posterior samples for the given data using the trained
            PyroVelocity model.
        compute_statistics_from_posterior_samples:
            Estimate statistics from posterior samples and add them to the
            `posterior_samples` dictionary.
        save_pyrovelocity_data:
            Saves the PyroVelocity data to a pickle file.
        save_model:
            Saves the Pyro-Velocity model to a directory.
        load_model:
            Load the model from a directory with the same structure as that produced
            by the save method.
    """

    def __init__(
        self,
        adata: AnnData,
        input_type: str = "raw",
        shared_time: bool = True,
        model_type: str = "auto",
        guide_type: str = "auto",
        likelihood: str = "Poisson",
        t_scale_on: bool = False,
        plate_size: int = 2,
        latent_factor: str = "none",
        latent_factor_operation: str = "selection",
        inducing_point_size: int = 0,
        latent_factor_size: int = 0,
        include_prior: bool = False,
        use_gpu: str = "auto",
        init: bool = False,
        num_aux_cells: int = 0,
        only_cell_times: bool = True,
        decoder_on: bool = False,
        add_offset: bool = False,
        correct_library_size: Union[bool, str] = True,
        cell_specific_kinetics: Optional[str] = None,
        kinetics_num: Optional[int] = None,
    ) -> None:
        """
        PyroVelocity class for estimating RNA velocity and related tasks.

        Args:
            adata (AnnData): An AnnData object containing the gene expression data.
            input_type (str, optional): Type of input data. Can be "raw", "knn", or "raw_cpm". Defaults to "raw".
            shared_time (bool, optional): Whether to use shared time. Defaults to True.
            model_type (str, optional): Type of model to use. Defaults to "auto".
            guide_type (str, optional): Type of guide to use. Defaults to "auto".
            likelihood (str, optional): Type of likelihood to use. Defaults to "Poisson".
            t_scale_on (bool, optional): Whether to use t_scale. Defaults to False.
            plate_size (int, optional): Size of the plate. Defaults to 2.
            latent_factor (str, optional): Type of latent factor. Defaults to "none".
            latent_factor_operation (str, optional): Operation to perform on the latent factor. Defaults to "selection".
            inducing_point_size (int, optional): Size of inducing points. Defaults to 0.
            latent_factor_size (int, optional): Size of latent factors. Defaults to 0.
            include_prior (bool, optional): Whether to include prior information. Defaults to False.
            use_gpu (Union[bool, int], optional): Whether and which GPU to use. Defaults to 0. Can be False.
            init (bool, optional): Whether to initialize the model. Defaults to False.
            num_aux_cells (int, optional): Number of auxiliary cells. Defaults to 0.
            only_cell_times (bool, optional): Whether to use only cell times. Defaults to True.
            decoder_on (bool, optional): Whether to use decoder. Defaults to False.
            add_offset (bool, optional): Whether to add offset. Defaults to False.
            correct_library_size (Union[bool, str], optional): Whether to correct library size or method to correct. Defaults to True.
            cell_specific_kinetics (Optional[str], optional): Type of cell-specific kinetics. Defaults to None.
            kinetics_num (Optional[int], optional): Number of kinetics. Defaults to None.

        Examples:
            >>> # import necessary libraries
            >>> import numpy as np
            >>> import anndata
            >>> from pyrovelocity.utils import pretty_log_dict, print_anndata, generate_sample_data
            >>> from pyrovelocity.tasks.preprocess import copy_raw_counts
            >>> from pyrovelocity.models._velocity import PyroVelocity
            ...
            >>> # define fixtures
            >>> try:
            >>>     tmp = getfixture("tmp_path")
            >>> except NameError:
            >>>     import tempfile
            >>>     tmp = tempfile.TemporaryDirectory().name
            >>> doctest_model_path = str(tmp) + "/save_pyrovelocity_doctest_model"
            >>> print(doctest_model_path)
            ...
            >>> # setup sample data
            >>> n_obs = 10
            >>> n_vars = 5
            >>> adata = generate_sample_data(n_obs=n_obs, n_vars=n_vars)
            >>> copy_raw_counts(adata)
            >>> print_anndata(adata)
            >>> print(adata.X)
            >>> print(adata.layers['spliced'])
            >>> print(adata.layers['unspliced'])
            >>> print(adata.obs['u_lib_size_raw'])
            >>> print(adata.obs['s_lib_size_raw'])
            >>> PyroVelocity.setup_anndata(adata)
            ...
            >>> # train model with macroscopic validation set
            >>> model = PyroVelocity(adata)
            >>> model.train(max_epochs=5, train_size=0.8, valid_size=0.2, use_gpu="auto")
            >>> posterior_samples = model.generate_posterior_samples(model.adata, num_samples=30)
            >>> print(posterior_samples.keys())
            >>> assert isinstance(posterior_samples, dict), f"Expected a dictionary, got {type(posterior_samples)}"
            >>> posterior_samples_log = pretty_log_dict(posterior_samples)
            >>> model.save_model(doctest_model_path, overwrite=True)
            >>> model = PyroVelocity.load_model(doctest_model_path, adata, use_gpu="auto")
            ...
            >>> # train model with default parameters
            >>> model = PyroVelocity(adata)
            >>> model.train_faster(max_epochs=5, use_gpu="auto")
            >>> model.save_model(doctest_model_path, overwrite=True)
            >>> model = PyroVelocity.load_model(doctest_model_path, adata, use_gpu="auto")
            >>> posterior_samples = model.generate_posterior_samples(model.adata, num_samples=30)
            >>> posterior_samples_log = pretty_log_dict(posterior_samples)
            >>> print(posterior_samples.keys())
            ...
            >>> # train model with specified batch size
            >>> model = PyroVelocity(adata)
            >>> model.train_faster_with_batch(batch_size=24, max_epochs=5, use_gpu="auto")
            >>> model.save_model(doctest_model_path, overwrite=True)
            >>> model = PyroVelocity.load_model(doctest_model_path, adata, use_gpu="auto")
            >>> posterior_samples = model.generate_posterior_samples(model.adata, num_samples=30)
            >>> posterior_samples_log = pretty_log_dict(posterior_samples)
            >>> print(posterior_samples.keys())
            ...
            >>> # If running from an interactive session, the temporary directory
            >>> # can be inspected to review the saved model files. When run as a
            >>> # doctest it is automatically cleaned up after the test completes.
            >>> print(f"Output located in {doctest_model_path}")
        """
        self.use_gpu = use_gpu
        self.cell_specific_kinetics = cell_specific_kinetics
        self.k = kinetics_num
        if input_type == "knn":
            layers = ["Mu", "Ms"]
            assert likelihood in {"Normal", "LogNormal"}
            assert "Mu" in adata.layers
        elif input_type == "raw_cpm":
            layers = ["unspliced", "spliced"]
            assert likelihood in {"Normal", "LogNormal"}
        else:
            layers = ["raw_unspliced", "raw_spliced"]
            assert likelihood != "Normal"

        self.layers = layers
        self.input_type = input_type

        super().__init__(adata)
        # TODO: remove unused code
        # from pyrovelocity.utils import init_with_all_cells
        # if init:
        #     initial_values = init_with_all_cells(
        #         self.adata,
        #         input_type,
        #         shared_time,
        #         latent_factor,
        #         latent_factor_size,
        #         plate_size,
        #         num_aux_cells=num_aux_cells,
        #     )
        # else:
        initial_values = {}
        logger.info(self.summary_stats)
        self.module = VelocityModule(
            self.summary_stats["n_cells"],
            self.summary_stats["n_vars"],
            model_type=model_type,
            guide_type=guide_type,
            likelihood=likelihood,
            shared_time=shared_time,
            t_scale_on=t_scale_on,
            plate_size=plate_size,
            latent_factor=latent_factor,
            latent_factor_operation=latent_factor_operation,
            latent_factor_size=latent_factor_size,
            inducing_point_size=inducing_point_size,
            include_prior=include_prior,
            use_gpu=use_gpu,
            num_aux_cells=num_aux_cells,
            only_cell_times=only_cell_times,
            decoder_on=decoder_on,
            add_offset=add_offset,
            correct_library_size=correct_library_size,
            cell_specific_kinetics=cell_specific_kinetics,
            kinetics_num=self.k,
            **initial_values,
        )
        self.num_cells = self.module.num_cells
        self._model_summary_string = """
        RNA velocity Pyro model with parameters:
        """
        self.init_params_ = self._get_init_params(locals())
        logger.info("Model initialized")

    def train(self, **kwargs):
        """
        Trains the PyroVelocity model using the provided data and configuration.

        The method leverages the Pyro library to train the model using the underlying
        data. It relies on the `VelocityTrainingMixin` to define the training logic.

        Args:

            **kwargs : dict, optional
                Additional keyword arguments to be passed to the underlying train method
                provided by the `VelocityTrainingMixin`.
        """
        pyro.enable_validation(True)
        super().train(**kwargs)

    def enum_parallel_predict(self):
        """work for parallel enumeration"""
        return

    @classmethod
    def setup_anndata(cls, adata: AnnData, *args, **kwargs):
        """
        Set up AnnData object for compatibility with the scvi-tools
        model training interface.

        Args:
            adata (AnnData): Anndata object to be used in model training.
        """
        setup_method_args = cls._get_setup_method_args(**locals())

        adata.obs["u_lib_size"] = np.log(
            adata.obs["u_lib_size_raw"].astype(float) + 1e-6
        )
        adata.obs["s_lib_size"] = np.log(
            adata.obs["s_lib_size_raw"].astype(float) + 1e-6
        )

        adata.obs["u_lib_size_mean"] = adata.obs["u_lib_size"].mean()
        adata.obs["s_lib_size_mean"] = adata.obs["s_lib_size"].mean()
        adata.obs["u_lib_size_scale"] = adata.obs["u_lib_size"].std()
        adata.obs["s_lib_size_scale"] = adata.obs["s_lib_size"].std()
        adata.obs["ind_x"] = np.arange(adata.n_obs).astype("int64")

        anndata_fields = [
            LayerField("U", "raw_unspliced", is_count_data=True),
            LayerField("X", "raw_spliced", is_count_data=True),
            NumericalObsField("u_lib_size", "u_lib_size"),
            NumericalObsField("s_lib_size", "s_lib_size"),
            NumericalObsField("u_lib_size_mean", "u_lib_size_mean"),
            NumericalObsField("s_lib_size_mean", "s_lib_size_mean"),
            NumericalObsField("u_lib_size_scale", "u_lib_size_scale"),
            NumericalObsField("s_lib_size_scale", "s_lib_size_scale"),
            NumericalObsField("ind_x", "ind_x"),
        ]
        adata_manager = AnnDataManager(
            fields=anndata_fields, setup_method_args=setup_method_args
        )
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    def generate_posterior_samples(
        self,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = None,
        num_samples: int = 100,
    ) -> Dict[str, ndarray]:
        """
        Generates posterior samples for the given data using the trained
        PyroVelocity model.

        The method generates posterior samples by running the trained model on the
        provided data and returns a dictionary containing samples for each parameter.

        Args:
            adata (AnnData, optional): Anndata object containing the data for which posterior samples
                are to be computed. If not provided, the anndata used to initialize the model will be used.
            indices (Sequence[int], optional): Indices of cells in `adata` for which the posterior
                samples are to be computed.
            batch_size (int, optional): The size of the mini-batches used during computation.
                If not provided, the entire dataset will be used.
            num_samples (int, default: 100): The number of posterior samples to compute for each parameter.

        Returns:
            Dict[str, ndarray]: A dictionary containing the posterior samples for each parameter.
        """
        self.module.eval()
        predictive = self.module.create_predictive(
            model=pyro.poutine.uncondition(self.module.model),
            num_samples=num_samples,
        )

        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size
        )

        with torch.no_grad(), pyro.poutine.mask(mask=False):
            posterior_samples = []
            for tensor in scdl:
                args, kwargs = self.module._get_fn_args_from_batch(tensor)
                posterior_sample = {
                    k: v.cpu().numpy()
                    for k, v in predictive(*args, **kwargs).items()
                }
                posterior_samples.append(posterior_sample)
            samples = {}
            for k in posterior_samples[0].keys():
                if k in [
                    "ut_norm",
                    "st_norm",
                    "time_constraint",
                ]:
                    continue

                if "aux" in k:
                    samples[k] = posterior_samples[0][k]
                elif posterior_samples[0][k].shape[-2] == 1:
                    samples[k] = posterior_samples[0][k]
                    if k == "kinetics_prob":
                        samples[k] = np.concatenate(
                            [
                                posterior_samples[j][k]
                                for j in range(len(posterior_samples))
                            ],
                            axis=-3,
                        )
                else:
                    samples[k] = np.concatenate(
                        [
                            posterior_samples[j][k]
                            for j in range(len(posterior_samples))
                        ],
                        axis=-2,
                    )

                logger.debug(k, "before", sys.getsizeof(samples[k]))
        self.num_samples = num_samples
        return samples

    def get_mlflow_logs(self):
        return

    def compute_statistics_from_posterior_samples(
        self,
        adata: AnnData,
        posterior_samples: Dict[str, ndarray],
        vector_field_basis: str = "umap",
        ncpus_use: int = 1,
    ) -> Dict[str, ndarray]:
        """
        Estimate statistics from posterior samples and add them to the
        `posterior_samples` dictionary. The names of the statistics incorporated into
        the dictionary are:

        - `gene_ranking`
        - `original_spaces_embeds_magnitude`
        - `genes`
        - `vector_field_posterior_samples`
        - `vector_field_posterior_mean`
        - `fdri`
        - `embeds_magnitude`
        - `embeds_angle`
        - `ut_mean`
        - `st_mean`
        - `pca_vector_field_posterior_samples`
        - `pca_embeds_angle`
        - `pca_fdri`

        The following data are removed from the `posterior_samples` dictionary:

        - `u`
        - `s`
        - `ut`
        - `st`

        Each of these sets requires further documentation.

        Args:
            adata (AnnData): Anndata object containing the data for which posterior samples
                were computed.
            posterior_samples (Dict[str, ndarray]): Dictionary containing the posterior samples
                for each parameter.
            vector_field_basis (str, optional): Basis for the vector field. Defaults to "umap".
            ncpus_use (int, optional): Number of CPUs to use for computation. Defaults to 1.

        Returns:
            Dict[str, ndarray]: Dictionary containing the posterior samples with added statistics.
        """
        if ("u_scale" in posterior_samples) and (
            "s_scale" in posterior_samples
        ):
            scale = posterior_samples["u_scale"] / posterior_samples["s_scale"]
        elif ("u_scale" in posterior_samples) and not (
            "s_scale" in posterior_samples
        ):
            scale = posterior_samples["u_scale"]
        else:
            scale = 1
        original_spaces_velocity_samples = (
            posterior_samples["beta"] * posterior_samples["ut"] / scale
            - posterior_samples["gamma"] * posterior_samples["st"]
        )
        original_spaces_embeds_magnitude = np.sqrt(
            (original_spaces_velocity_samples**2).sum(axis=-1)
        )

        (
            vector_field_posterior_samples,
            embeds_radian,
            fdri,
        ) = vector_field_uncertainty(
            adata,
            posterior_samples,
            basis=vector_field_basis,
            n_jobs=ncpus_use,
        )
        embeds_magnitude = np.sqrt(
            (vector_field_posterior_samples**2).sum(axis=-1)
        )

        mlflow.log_metric(
            "FDR_sig_frac", round((fdri < 0.05).sum() / fdri.shape[0], 3)
        )
        mlflow.log_metric("FDR_HMP", harmonic_mean(fdri))

        compute_mean_vector_field(
            posterior_samples=posterior_samples,
            adata=adata,
            basis=vector_field_basis,
            n_jobs=ncpus_use,
        )

        vector_field_posterior_mean = adata.obsm[
            f"velocity_pyro_{vector_field_basis}"
        ]

        gene_ranking, genes = compute_volcano_data(
            [posterior_samples], [adata], time_correlation_with="st"
        )
        gene_ranking = (
            gene_ranking.sort_values("mean_mae", ascending=False)
            .head(300)
            .sort_values("time_correlation", ascending=False)
        )
        posterior_samples["gene_ranking"] = gene_ranking
        posterior_samples[
            "original_spaces_embeds_magnitude"
        ] = original_spaces_embeds_magnitude
        posterior_samples["genes"] = genes
        posterior_samples[
            "vector_field_posterior_samples"
        ] = vector_field_posterior_samples
        posterior_samples[
            "vector_field_posterior_mean"
        ] = vector_field_posterior_mean
        posterior_samples["fdri"] = fdri
        posterior_samples["embeds_magnitude"] = embeds_magnitude
        posterior_samples["embeds_angle"] = embeds_radian
        posterior_samples["ut_mean"] = posterior_samples["ut"].mean(0).squeeze()
        posterior_samples["st_mean"] = posterior_samples["st"].mean(0).squeeze()

        (
            pca_vector_field_posterior_samples,
            pca_embeds_radian,
            pca_fdri,
        ) = vector_field_uncertainty(
            adata,
            posterior_samples,
            basis="pca",
            n_jobs=ncpus_use,
        )
        posterior_samples[
            "pca_vector_field_posterior_samples"
        ] = pca_vector_field_posterior_samples
        posterior_samples["pca_embeds_angle"] = pca_embeds_radian
        posterior_samples["pca_fdri"] = pca_fdri

        del posterior_samples["u"]
        del posterior_samples["s"]
        del posterior_samples["ut"]
        del posterior_samples["st"]
        return posterior_samples

    @beartype
    def save_pyrovelocity_data(
        self,
        posterior_samples: Dict[str, ndarray],
        pyrovelocity_data_path: os.PathLike | str,
    ):
        """
        Save the PyroVelocity data to a pickle file.

        Args:
            posterior_samples (Dict[str, ndarray]): Dictionary containing the posterior samples
            pyrovelocity_data_path (os.PathLike | str): Path to save the PyroVelocity data.
        """
        with open(pyrovelocity_data_path, "wb") as f:
            pickle.dump(posterior_samples, f)
        for k in posterior_samples:
            logger.debug(k, "after", sys.getsizeof(posterior_samples[k]))

    def save_model(
        self,
        dir_path: str,
        prefix: Optional[str] = None,
        overwrite: bool = True,
        save_anndata: bool = False,
        **anndata_write_kwargs,
    ) -> None:
        """
        Save the Pyro-Velocity model to a directory.

        Dispatches to the `save` method of the inherited `BaseModelClass` which
        calls `torch.save` on a model state dictionary, variable names, and user
        attributes.

        Args:
            dir_path (str): Path to the directory where the model will be saved.
            prefix (Optional[str], optional): Prefix to add to the saved files. Defaults to None.
            overwrite (bool, optional): Whether to overwrite existing files. Defaults to True.
            save_anndata (bool, optional): Whether to save the AnnData object. Defaults to False.
        """
        super().save(
            dir_path, prefix, overwrite, save_anndata, **anndata_write_kwargs
        )
        pyro.get_param_store().save(
            os.path.join(dir_path, "param_store_test.pt")
        )

    @classmethod
    def load_model(
        cls,
        dir_path: str,
        adata: Optional[AnnData] = None,
        use_gpu: str = "auto",
        prefix: Optional[str] = None,
        backup_url: Optional[str] = None,
    ) -> BaseModelClass:
        """
        Load the model from a directory with the same structure as that produced
        by the save method.

        Args:
            dir_path (str): Path to the directory where the model is saved.
            adata (Optional[AnnData], optional): Anndata object to load into the model. Defaults to None.
            use_gpu (str, optional): Whether and which GPU to use. Defaults to "auto".
            prefix (Optional[str], optional): Prefix to add to the saved files. Defaults to None.
            backup_url (Optional[str], optional): URL to download the model from. Defaults to None.

        Raises:
            RuntimeError: If the model is not an instance of PyroBaseModuleClass.

        Returns:
            PyroVelocity: The loaded PyroVelocity model.
        """
        load_adata = adata is None
        _accelerator, _devices, device = parse_device_args(
            accelerator=use_gpu, return_device="torch"
        )
        logger.info(
            f"\nLoading model with:\n"
            f"\taccelerator: {_accelerator}\n"
            f"\tdevices: {_devices}\n"
            f"\tdevice: {device}\n"
        )

        (
            attr_dict,
            var_names,
            model_state_dict,
            new_adata,
        ) = _load_saved_files(
            dir_path,
            load_adata,
            map_location=device,
            prefix=prefix,
            backup_url=backup_url,
        )

        adata = new_adata if new_adata is not None else adata

        _validate_var_names(adata, var_names)

        registry = attr_dict.pop("registry_")
        method_name = registry.get(_SETUP_METHOD_NAME, "setup_anndata")
        getattr(cls, method_name)(
            adata, source_registry=registry, **registry[_SETUP_ARGS_KEY]
        )

        model = _initialize_model(cls, adata, attr_dict)

        for attr, val in attr_dict.items():
            setattr(model, attr, val)

        pyro.clear_param_store()
        old_history = model.history_
        try:
            model.module.load_state_dict(model_state_dict)
        except RuntimeError as err:
            if not isinstance(model.module, PyroBaseModuleClass):
                raise err
            logger.info(
                "Preparing underlying `PyroBaseModuleClass` module for load"
            )
            try:
                model.train(max_epochs=1, max_steps=1)
            except Exception:
                model.train(
                    max_epochs=1,
                    max_steps=1,
                    batch_size=adata.shape[0],
                    train_size=0.8,
                    valid_size=0.2,
                )
            model.module.load_state_dict(model_state_dict)

        model.history_ = old_history
        model.to_device(device)
        model.module.eval()
        model._validate_anndata(adata)
        pyro.get_param_store().load(
            os.path.join(dir_path, "param_store_test.pt"),
            map_location=device,
        )
        return model



import matplotlib
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from pyrovelocity.logging import configure_logging


__all__ = ["set_colorbar", "set_font_size"]

logger = configure_logging(__name__)


def set_font_size(size: int):
    matplotlib.rcParams.update({"font.size": size})


def set_colorbar(
    smp,
    ax,
    orientation="vertical",
    labelsize=None,
    fig=None,
    position="right",
    rainbow=False,
):
    if position == "right" and (not rainbow):
        cax = inset_axes(ax, width="2%", height="30%", loc=4, borderpad=0)
        cb = fig.colorbar(smp, orientation=orientation, cax=cax)
    else:
        divider = make_axes_locatable(ax)
        cax = divider.append_axes(position, size="8%", pad=0.08)
        cb = fig.colorbar(smp, cax=cax, orientation=orientation, shrink=0.4)

    cb.ax.tick_params(labelsize=labelsize)

    # TODO: remove cb.draw_all()
    #
    # MatplotlibDeprecationWarning: The draw_all function was deprecated in
    # Matplotlib 3.6 and will be removed two minor releases later. Use
    # fig.draw_without_rendering() instead. cbar.draw_all()
    #
    # draw_all is not required with cb.solids.set_alpha(1)
    # https://matplotlib.org/stable/api/colorbar_api.html#matplotlib.colorbar.Colorbar.set_alpha
    cb.solids.set_alpha(1)
    # cb.set_alpha(1)
    # cb.draw_all()

    cb.locator = MaxNLocator(nbins=2, integer=True)

    if position == "left":
        cb.ax.yaxis.set_ticks_position("left")
    cb.update_ticks()



import os

import anyio
import appdirs
import httpx
from anyio import Semaphore
from beartype import beartype
from beartype.typing import List
from beartype.typing import Tuple
from diskcache import Cache
from returns.result import Failure
from returns.result import Result
from returns.result import Success

from pyrovelocity.logging import configure_logging


__all__ = ["fetch_gene_sequences"]

logger = configure_logging(__name__)

CONCURRENCY_LIMIT = 100


@beartype
async def fetch_sequence(
    gene_id: str,
    cache: Cache,
    client: httpx.AsyncClient,
    semaphore: Semaphore,
) -> Result[str, str]:
    """
    Fetch the genomic sequence for a given gene or transcript ID from Ensembl.

    Results are cached to avoid redundant requests.

    Args:
        gene_id (str): Ensembl gene or transcript ID.
        cache (Cache): DiskCache object for caching results.

    Returns:
        Result[str, str]: A result object encapsulating success or failure.
    """
    async with semaphore:
        if gene_id in cache:
            logger.info(f"Retrieved from cache: {gene_id}")
            return Success(cache[gene_id])

        url = f"https://rest.ensembl.org/sequence/id/{gene_id}?type=genomic"
        headers = {"Content-Type": "text/x-fasta"}
        response = await client.get(url, headers=headers)

        if response.status_code == 200:
            rate_limit_remaining = int(
                response.headers.get("x-ratelimit-remaining", 0)
            )
            if rate_limit_remaining % 1000 == 0:
                logger.warning(f"Rate limit remaining: {rate_limit_remaining}")

            sequence = response.text
            cache[gene_id] = sequence
            logger.info(f"Fetched and cached: {gene_id}")
            return Success(sequence)
        else:
            error_message = f"Failed to fetch {gene_id}: {response.status_code}"
            logger.error(error_message)
            return Failure(error_message)


async def main(gene_id_list: List[str], cache: Cache) -> List[Result[str, str]]:
    results = []
    semaphore = Semaphore(CONCURRENCY_LIMIT)
    async with httpx.AsyncClient(
        timeout=10.0
    ) as client, anyio.create_task_group() as tg:
        for gene_id in gene_id_list:
            tg.start_soon(
                fetch_and_log_sequence,
                gene_id,
                cache,
                client,
                results,
                semaphore,
            )
    return results


@beartype
async def fetch_and_log_sequence(
    gene_id: str,
    cache: Cache,
    client: httpx.AsyncClient,
    results: List,
    semaphore: Semaphore,
):
    logger.info(f"Starting fetch for {gene_id}")
    result = await fetch_sequence(gene_id, cache, client, semaphore)
    results.append(result)
    logger.info(f"Completed fetch for {gene_id}")


@beartype
def process_results(
    results: List[Result[str, str]]
) -> Tuple[List[str], List[str]]:
    """
    Separates the successful results from the failures.

    Args:
        results (List[Result[str, str]]): List of Result objects.

    Returns:
        Tuple[List[str], List[str]]:
            A tuple containing a list of successful sequences and
            a list of error messages.
    """
    successes = [
        result.unwrap() for result in results if isinstance(result, Success)
    ]
    failures = [
        result.failure() for result in results if isinstance(result, Failure)
    ]
    return (successes, failures)


@beartype
def fetch_gene_sequences(
    gene_id_list: List[str],
    cache_path: str = "./.ensembl",
    fasta_file_name: str = "genomic_sequences",
):
    """

    Example:
    >>> # xdoctest: +SKIP
    >>> gene_id_list = [
    ...   'ENSG00000121410',
    ...   'ENSG00000175899',
    ...   'ENSG00000166535',
    ...   # 'ENSG00000128274',
    >>> ]
    >>> print(gene_id_list)
    >>> sequences, _ = fetch_gene_sequences(gene_id_list)
    >>> print(sequences)
    """
    cache_dir = appdirs.user_cache_dir("pyrovelocity")
    os.makedirs(cache_dir, exist_ok=True)
    fasta_file_path = os.path.join(cache_dir, f"{fasta_file_name}.fasta")

    cache = Cache(cache_path)
    try:
        sequences = anyio.run(main, gene_id_list, cache, backend="trio")
        successes, failures = process_results(sequences)

        with open(fasta_file_path, "w") as fasta_file:
            for sequence in successes:
                fasta_file.write(f"{sequence}")

        logger.info(f"Sequences saved to {fasta_file_path}")
    finally:
        cache.close()

    return successes, failures



from typing import Tuple

import torch
from beartype import beartype
from torch import Tensor


@beartype
def mrna_dynamics(
    tau: Tensor,
    u0: Tensor,
    s0: Tensor,
    alpha: Tensor,
    beta: Tensor,
    gamma: Tensor,
) -> Tuple[Tensor, Tensor]:
    """
    Computes the mRNA dynamics given temporal coordinate, parameter values, and
    initial conditions.

    `st_gamma_equals_beta` for the case where the gamma parameter is equal
    to the beta parameter is taken from Equation 2.12 of

    > Li T, Shi J, Wu Y, Zhou P. On the mathematics of RNA velocity I:
    Theoretical analysis. CSIAM Transactions on Applied Mathematics. 2021;2:
    1–55. doi:[10.4208/csiam-am.so-2020-0001](https://doi.org/10.4208/csiam-am.so-2020-0001)

    Args:
        tau (Tensor): Time points.
        u0 (Tensor): Initial value of u.
        s0 (Tensor): Initial value of s.
        alpha (Tensor): Alpha parameter.
        beta (Tensor): Beta parameter.
        gamma (Tensor): Gamma parameter.

    Returns:
        Tuple[Tensor, Tensor]: Tuple containing the final values of u and s.

    Examples:
        >>> import torch
        >>> tau = torch.tensor(2.0)
        >>> u0 = torch.tensor(1.0)
        >>> s0 = torch.tensor(0.5)
        >>> alpha = torch.tensor(0.5)
        >>> beta = torch.tensor(0.4)
        >>> gamma = torch.tensor(0.3)
        >>> mrna_dynamics(tau, u0, s0, alpha, beta, gamma)
        (tensor(1.1377), tensor(0.9269))
    """
    expu, exps = torch.exp(-beta * tau), torch.exp(-gamma * tau)

    expus = (alpha - u0 * beta) * inv(gamma - beta) * (exps - expu)

    ut = u0 * expu + alpha / beta * (1 - expu)
    st = s0 * exps + alpha / gamma * (1 - exps) + expus
    st_gamma_equals_beta = (
        s0 * expu + alpha / beta * (1 - expu) - (alpha - beta * u0) * tau * expu
    )
    st = torch.where(torch.isclose(gamma, beta), st_gamma_equals_beta, st)

    return ut, st


@beartype
def inv(x: Tensor) -> Tensor:
    """
    Computes the element-wise reciprocal of a tensor.

    Args:
        x (Tensor): Input tensor.

    Returns:
        Tensor: Tensor with element-wise reciprocal of x.

    Examples:
        >>> import torch
        >>> x = torch.tensor([2., 4., 0.5])
        >>> inv(x)
        tensor([0.5000, 0.2500, 2.0000])
    """
    return x.reciprocal()



import os
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from typing import List

from beartype import beartype
from google.oauth2 import service_account
from googleapiclient.discovery import Resource
from googleapiclient.discovery import build
from googleapiclient.http import MediaFileUpload

from pyrovelocity.logging import configure_logging


logger = configure_logging(__name__)


@beartype
def authenticate_gdrive(
    service_account_file: str = "service-account.json",
    scopes: List[str] = ["https://www.googleapis.com/auth/drive"],
) -> Resource:
    """
    Authenticate with Google Drive API using a service account.

    Args:
        service_account_file (str, optional): Path to the service account JSON file. Defaults to "service-account.json".
        scopes (List[str], optional): List of scopes to request during authentication. Defaults to ["https://www.googleapis.com/auth/drive"].

    Returns:
        Resource: An authenticated Google Drive API service object.

    Examples:
        >>> authenticate_gdrive() # xdoctest: +SKIP
    """
    credentials = service_account.Credentials.from_service_account_file(
        service_account_file, scopes=scopes
    )
    service = build("drive", "v3", credentials=credentials)
    return service


def _get_mimetype(file_path):
    """Get the mimetype based on file extension"""
    import mimetypes

    mimetype, _ = mimetypes.guess_type(file_path)
    return mimetype or "application/octet-stream"


@beartype
def upload_file_to_drive(
    service: Resource, file_path: str, folder_id: str
) -> str:
    """
    Uploads an file file to a specified Google Drive folder.

    Args:
        service: An authenticated Google Drive API service object.
        file_path (str): Path to the file to be uploaded.
        folder_id (str): The ID of the Google Drive folder where the file will be uploaded.

    Returns:
        str: The ID of the uploaded file.

    Examples:
        >>> service = authenticate_gdrive("pyrovelocity-drive-sa.json") # xdoctest: +SKIP
        >>> upload_file_to_drive(service, "docs/_static/logo.png", "1-0y0Gw1HIC3o8ooRcHWO-PHuPPtPKYTt") # xdoctest: +SKIP
    """
    file_metadata = {
        "name": os.path.basename(file_path),
        "parents": [folder_id],
    }
    media = MediaFileUpload(file_path, mimetype=_get_mimetype(file_path))
    file = (
        service.files()
        .create(body=file_metadata, media_body=media, fields="id")
        .execute()
    )
    return file.get("id")


@beartype
def create_folder_in_drive(
    service: Resource, folder_name: str, parent_folder_id: str
) -> str:
    """
    Creates a folder in Google Drive.

    Args:
        service: An authenticated Google Drive API service object.
        folder_name (str): Name of the folder to be created.
        parent_folder_id (str): The ID of the parent Google Drive folder.

    Returns:
        str: The ID of the created folder.
    """
    file_metadata = {
        "name": folder_name,
        "mimeType": "application/vnd.google-apps.folder",
        "parents": [parent_folder_id],
    }
    folder = service.files().create(body=file_metadata, fields="id").execute()
    return folder.get("id")


@beartype
def create_drive_folder_structure(
    service: Resource,
    folder_path: str,
    parent_folder_id: str,
    top_folder_name: str,
) -> dict:
    top_folder_id = create_folder_in_drive(
        service, top_folder_name, parent_folder_id
    )
    folder_mapping = {folder_path: top_folder_id}

    for root, dirs, _ in os.walk(folder_path):
        for dir_name in dirs:
            local_dir_path = os.path.join(root, dir_name)
            local_parent_dir = os.path.dirname(local_dir_path)
            drive_parent_id = folder_mapping[local_parent_dir]
            drive_folder_id = create_folder_in_drive(
                service, dir_name, drive_parent_id
            )
            folder_mapping[local_dir_path] = drive_folder_id

    return folder_mapping


@beartype
def upload_files_to_drive(
    service: Resource,
    folder_path: str,
    folder_mapping: dict,
    max_workers: int = 7,
):
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_file = {}
        for root, _, files in os.walk(folder_path):
            for file_name in files:
                local_file_path = os.path.join(root, file_name)
                drive_folder_id = folder_mapping[root]
                future = executor.submit(
                    upload_file_to_drive,
                    service,
                    local_file_path,
                    drive_folder_id,
                )
                future_to_file[future] = local_file_path

        for future in as_completed(future_to_file):
            file_path = future_to_file[future]
            try:
                file_id = future.result()
                logger.info(f"Uploaded '{file_path}' with File ID: {file_id}")
            except Exception as e:
                logger.error(f"Failed to upload '{file_path}'. Error: {e}")


@beartype
def upload_folder(
    service: Resource,
    folder_path: str,
    parent_folder_id: str,
    top_folder_name: str,
    max_workers: int = 7,
):
    """
    Uploads an entire directory to Google Drive.

    Args:
        service: An authenticated Google Drive API service object.
        folder_path (str): The local directory path to upload.
        parent_folder_id (str): The ID of the parent Google Drive folder.
        top_folder_name (str): The name of the top-level folder in Google Drive.
        max_workers (int, optional): The maximum number of threads to use for uploading files. Defaults to 7.

    Examples:
        >>> service = authenticate_gdrive("pyrovelocity-drive-sa.json") # xdoctest: +SKIP
        >>> upload_folder(service, "docs/_static/", "1-0y0Gw1HIC3o8ooRcHWO-PHuPPtPKYTt", "test_00/") # xdoctest: +SKIP
    """
    folder_mapping = create_drive_folder_structure(
        service, folder_path, parent_folder_id, top_folder_name
    )
    upload_files_to_drive(service, folder_path, folder_mapping, max_workers)



import pprint

import pandas as pd
from beartype import beartype


__all__ = ["generate_dataframe_fixture_code"]


@beartype
def generate_dataframe_fixture_code(
    df: pd.DataFrame,
    df_name: str = "df",
) -> str:
    """
    Generate a code string to create a pandas DataFrame from a given DataFrame.

    Args:
        df (pd.DataFrame): A pandas DataFrame to generate code for.
        df_name (str, optional): Name of the DataFrame in the code.
            Defaults to "df".

    Raises:
        ValueError: If the input is not a pandas DataFrame

    Returns:
        str: String containing python code to explicitly generate a pandas
            DataFrame from the input DataFrame.
    """
    if not isinstance(df, pd.DataFrame):
        raise ValueError("Input must be a pandas DataFrame")

    data_dict_str = df.to_dict(orient="list")

    pp = pprint.PrettyPrinter(indent=4)
    data_dict_pretty_str = pp.pformat(data_dict_str)

    code_str = f"""import pandas as pd

{df_name} = pd.DataFrame(
    {data_dict_pretty_str}
)
"""
    return code_str



from os import PathLike
from typing import Dict

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import scvelo as scv
import seaborn as sns
from anndata import AnnData
from beartype import beartype
from matplotlib.figure import FigureBase
from scipy.stats import spearmanr

from pyrovelocity.plots._common import set_colorbar
from pyrovelocity.plots._common import set_font_size


__all__ = [
    "plot_posterior_time",
    "plot_shared_time_uncertainty",
]


def plot_posterior_time(
    posterior_samples,
    adata,
    ax=None,
    fig=None,
    basis="umap",
    addition=True,
    position="left",
    s=3,
):
    if addition:
        sns.set_style("white")
        sns.set_context("paper", font_scale=1)
        set_font_size(7)
        plt.figure()
        plt.hist(posterior_samples["cell_time"].mean(0), bins=100, label="test")
        plt.xlabel("mean of cell time")
        plt.ylabel("frequency")
        plt.title("Histogram of cell time posterior samples")
        plt.legend()
    pos_mean_time = posterior_samples["cell_time"].mean(0)
    adata.obs["cell_time"] = pos_mean_time / pos_mean_time.max()

    if ax is None:
        fig, ax = plt.subplots(1, 1)
        fig.set_size_inches(2.36, 2)
    im = ax.scatter(
        adata.obsm[f"X_{basis}"][:, 0],
        adata.obsm[f"X_{basis}"][:, 1],
        s=s,
        alpha=0.4,
        c=adata.obs["cell_time"],
        cmap="inferno",
        linewidth=0,
    )
    set_colorbar(im, ax, labelsize=5, fig=fig, position=position)
    ax.axis("off")
    if "cytotrace" in adata.obs.columns:
        ax.set_title(
            "Pyro-Velocity shared time\ncorrelation with Cytotrace: %.2f"
            % (
                spearmanr(
                    adata.obs["cell_time"].values,
                    1 - adata.obs.cytotrace.values,
                )[0]
            ),
            fontsize=7,
        )
    else:
        ax.set_title("Pyro-Velocity shared time\n", fontsize=7)


@beartype
def plot_shared_time_uncertainty(
    posterior_samples: Dict[str, np.ndarray],
    adata: AnnData,
    vector_field_basis: str,
    shared_time_plot: PathLike | str,
) -> FigureBase:
    cell_time_mean = posterior_samples["cell_time"].mean(0).flatten()
    cell_time_std = posterior_samples["cell_time"].std(0).flatten()
    adata.obs["shared_time_std"] = cell_time_std
    adata.obs["shared_time_mean"] = cell_time_mean

    set_font_size(7)
    fig, ax = plt.subplots(1, 3)
    fig.set_size_inches(10, 3)
    ax = ax.flatten()

    ax[0].hist(cell_time_std / cell_time_mean, bins=100)
    ax[0].set_title("histogram of shared time CoV")
    ax_st = scv.pl.scatter(
        adata,
        c="shared_time_mean",
        ax=ax[1],
        show=False,
        cmap="inferno",
        fontsize=12,
        colorbar=True,
    )
    ax_cv = scv.pl.scatter(
        adata,
        c="shared_time_std",
        ax=ax[2],
        show=False,
        cmap="winter",
        fontsize=12,
        colorbar=True,
        title="shared time standard deviation",
    )
    ax_cv.set_xlabel("density estimate over 90th %")
    select = adata.obs["shared_time_std"] > np.quantile(
        adata.obs["shared_time_std"], 0.9
    )
    sns.kdeplot(
        x=adata.obsm[f"X_{vector_field_basis}"][:, 0][select],
        y=adata.obsm[f"X_{vector_field_basis}"][:, 1][select],
        ax=ax[2],
        levels=3,
        fill=False,
    )
    fig.tight_layout()
    for ext in ["", ".png"]:
        fig.savefig(
            f"{shared_time_plot}{ext}",
            facecolor=fig.get_facecolor(),
            bbox_inches="tight",
            edgecolor="none",
            dpi=300,
        )

    plt.close(fig)

    return fig



from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import scanpy as sc
import scvelo as scv
import sklearn
import umap
from anndata import AnnData
from astropy.stats import rayleightest
from beartype import beartype
from numpy import ndarray
from sklearn.pipeline import Pipeline

from pyrovelocity.analysis.cytotrace import compute_similarity2
from pyrovelocity.logging import configure_logging
from pyrovelocity.utils import ensure_numpy_array

__all__ = [
    "compute_mean_vector_field",
    "compute_volcano_data",
    "mae_per_gene",
    "vector_field_uncertainty",
]

logger = configure_logging(__name__)


def compute_mean_vector_field(
    posterior_samples,
    adata,
    basis="umap",
    n_jobs=1,
    spliced="spliced_pyro",
    raw=False,
):
    logger.info("Computing mean vector field")
    # scv.pp.neighbors(adata, use_rep="pca")
    if "X_pca" not in adata.obsm.keys():
        sc.pp.pca(
            data=adata,
            svd_solver="arpack",
        )
    sc.pp.neighbors(adata=adata, n_neighbors=30, use_rep="X_pca")

    adata.var["velocity_genes"] = True

    if spliced == "spliced_pyro":
        if raw:
            ut = posterior_samples["ut"]
            st = posterior_samples["st"]
            ut = ut / ut.sum(axis=-1, keepdims=True)
            st = st / st.sum(axis=-1, keepdims=True)
        else:
            ut = posterior_samples["ut"]
            st = posterior_samples["st"]
        adata.layers["spliced_pyro"] = st.mean(0).squeeze()
        # if ('u_scale' in posterior_samples) and ('s_scale' in posterior_samples)
        # TODO: two scale for Normal distribution
        if (
            "u_scale" in posterior_samples
        ):  # only one scale for Poisson distribution
            adata.layers["velocity_pyro"] = (
                ut * posterior_samples["beta"] / posterior_samples["u_scale"]
                - st * posterior_samples["gamma"]
            ).mean(0)
        else:
            if "beta_k" in posterior_samples:
                adata.layers["velocity_pyro"] = (
                    (
                        ut * posterior_samples["beta_k"]
                        - posterior_samples["st"] * posterior_samples["gamma_k"]
                    )
                    .mean(0)
                    .squeeze()
                )
            else:
                adata.layers["velocity_pyro"] = (
                    ut * posterior_samples["beta"]
                    - posterior_samples["st"] * posterior_samples["gamma"]
                ).mean(0)
        scv.tl.velocity_graph(
            adata, vkey="velocity_pyro", xkey="spliced_pyro", n_jobs=n_jobs
        )
    elif spliced in ["Ms"]:
        ut = adata.layers["Mu"]
        st = adata.layers["Ms"]
        if ("u_scale" in posterior_samples) and (
            "s_scale" in posterior_samples
        ):
            adata.layers["velocity_pyro"] = (
                ut
                * posterior_samples["beta"]
                / (posterior_samples["u_scale"] / posterior_samples["s_scale"])
                - st * posterior_samples["gamma"]
            ).mean(0)
        else:
            adata.layers["velocity_pyro"] = (
                ut * posterior_samples["beta"]
                - posterior_samples["st"] * posterior_samples["gamma"]
            ).mean(0)
        scv.tl.velocity_graph(
            adata, vkey="velocity_pyro", xkey="Ms", n_jobs=n_jobs
        )
    elif spliced in ["spliced"]:
        ut = adata.layers["unspliced"]
        st = adata.layers["spliced"]
        if ("u_scale" in posterior_samples) and (
            "s_scale" in posterior_samples
        ):
            adata.layers["velocity_pyro"] = (
                ut
                * posterior_samples["beta"]
                / (posterior_samples["u_scale"] / posterior_samples["s_scale"])
                - st * posterior_samples["gamma"]
            ).mean(0)
        else:
            adata.layers["velocity_pyro"] = (
                ut * posterior_samples["beta"]
                - posterior_samples["st"] * posterior_samples["gamma"]
            ).mean(0)
        scv.tl.velocity_graph(
            adata, vkey="velocity_pyro", xkey="spliced", n_jobs=n_jobs
        )

    scv.tl.velocity_embedding(adata, vkey="velocity_pyro", basis=basis)


@beartype
def compute_volcano_data(
    posterior_samples: List[Dict[str, ndarray]],
    adata: List[AnnData],
    time_correlation_with: str = "s",
    selected_genes: Optional[List[str]] = None,
    negative: bool = False,
) -> Tuple[pd.DataFrame, List[str]]:
    assert isinstance(posterior_samples, (tuple, list))
    assert isinstance(adata, (tuple, list))
    assert "s" in posterior_samples[0]
    assert "alpha" in posterior_samples[0]

    maes_list = []
    cors = []
    genes = []
    labels = []
    switching = []
    for p, ad, label in zip(posterior_samples, adata, ["train", "valid"]):
        print(label)
        for sample in range(p["alpha"].shape[0]):
            maes_list.append(
                mae_per_gene(
                    p["s"][sample].squeeze(),
                    ensure_numpy_array(ad.layers["raw_spliced"]),
                )
            )
            df_genes_cors = compute_similarity2(
                p[time_correlation_with][sample].squeeze(),
                p["cell_time"][sample].squeeze().reshape(-1, 1),
            )
            cors.append(df_genes_cors[0])
            genes.append(ad.var_names.values)
            labels.append([f"Poisson_{label}"] * len(ad.var_names.values))

    volcano_data = pd.DataFrame(
        {
            "mean_mae": np.hstack(maes_list),
            "label": np.hstack(labels),
            "time_correlation": np.hstack(cors),
            "genes": np.hstack(genes),
        }
    )
    volcano_data = volcano_data.groupby("genes").mean(
        ["mean_mae", "time_correlation"]
    )

    volcano_data.loc[:, "mean_mae_rank"] = volcano_data.mean_mae.rank(
        ascending=False
    )
    volcano_data.loc[
        :, "time_correlation_rank"
    ] = volcano_data.time_correlation.apply(abs).rank(ascending=False)
    volcano_data.loc[:, "rank_product"] = (
        volcano_data.mean_mae_rank * volcano_data.time_correlation_rank
    )

    if selected_genes is None:
        genes = (
            volcano_data.sort_values("mean_mae", ascending=False)
            .head(300)
            .sort_values("time_correlation", ascending=negative)
            .head(4)
            .index.tolist()
        )
    else:
        genes = selected_genes
    volcano_data.loc[:, "selected genes"] = 0
    volcano_data.loc[genes, "selected genes"] = 1

    return volcano_data, genes


def vector_field_uncertainty(
    adata: AnnData,
    posterior_samples: dict[str, ndarray],
    basis: str = "tsne",
    n_jobs: int = 1,
    denoised: bool = False,
) -> Tuple[ndarray, ndarray, ndarray]:
    """Run cosine similarity-based vector field across posterior samples"""

    logger.info("Estimating vector field uncertainty")
    # fig, ax = plt.subplots(10, 3)
    # fig.set_size_inches(16, 36)
    # ax = ax.flatten()
    v_map_all = []
    if ("u_scale" in posterior_samples) and (
        "s_scale" in posterior_samples
    ):  # Gaussian models
        scale = posterior_samples["u_scale"] / posterior_samples["s_scale"]
    elif ("u_scale" in posterior_samples) and not (
        "s_scale" in posterior_samples
    ):  # Poisson Model 2
        scale = posterior_samples["u_scale"]
    else:  # Poisson Model 1
        scale = 1

    if "beta_k" in posterior_samples:
        velocity_samples = (
            posterior_samples["ut"] * posterior_samples["beta_k"] / scale
            - posterior_samples["st"] * posterior_samples["gamma_k"]
        )
    else:
        velocity_samples = (
            posterior_samples["beta"] * posterior_samples["ut"] / scale
            - posterior_samples["gamma"] * posterior_samples["st"]
        )
    if denoised:
        projection = [
            (
                "PCA",
                sklearn.decomposition.PCA(random_state=99, n_components=50),
            ),
            ("UMAP", umap.UMAP(random_state=99, n_components=2)),
        ]
        pipelines = Pipeline(projection)
        expression = [posterior_samples["st"].mean(0)]
        pipelines.fit(expression[0])
        umap_orig = pipelines.transform(expression[0])
        adata.obsm["X_umap1"] = umap_orig
        joint_pcs = pipelines.steps[0][1].transform(expression[0])
        adata.obsm["X_pyropca"] = joint_pcs
        # scv.pp.neighbors(adata, use_rep="pyropca")
        sc.pp.neighbors(adata=adata, n_neighbors=30, use_rep="X_pyropca")
    else:
        # scv.pp.neighbors(adata, use_rep="pca")
        if "X_pca" not in adata.obsm.keys():
            sc.pp.pca(
                data=adata,
                svd_solver="arpack",
            )
        sc.pp.neighbors(adata=adata, n_neighbors=30, use_rep="X_pca")

    assert len(posterior_samples["st"].shape) == 3
    adata.var["velocity_genes"] = True
    for sample in range(posterior_samples["st"].shape[0]):
        adata.layers["spliced_pyro"] = posterior_samples["st"][sample]
        adata.layers["velocity_pyro"] = velocity_samples[sample]

        if basis == "pca":
            sc.pp.pca(
                data=adata,
                svd_solver="arpack",
            )
            scv.tl.velocity_embedding(
                adata,
                vkey="velocity_pyro",
                basis="pca",
                direct_pca_projection=True,
            )
        else:
            scv.tl.velocity_graph(
                adata, vkey="velocity_pyro", xkey="spliced_pyro", n_jobs=n_jobs
            )
            scv.tl.velocity_embedding(adata, vkey="velocity_pyro", basis=basis)
        v_map_all.append(adata.obsm[f"velocity_pyro_{basis}"])
    v_map_all = np.stack(v_map_all)
    embeds_radian = np.arctan2(v_map_all[:, :, 1], v_map_all[:, :, 0])
    from statsmodels.stats.multitest import multipletests

    rayleightest_pval = rayleightest(embeds_radian, axis=-2)
    _, fdri, _, _ = multipletests(rayleightest_pval, method="fdr_bh")
    return v_map_all, embeds_radian, fdri


def mae_per_gene(pred_counts: ndarray, true_counts: ndarray) -> ndarray:
    """Computes mean average error between counts and predicted probabilities."""
    error = np.abs(true_counts - pred_counts).sum(-2)
    total = np.clip(true_counts.sum(-2), 1, np.inf)
    return -np.array(error / total)


@beartype
def pareto_frontier_genes(
    volcano_data: pd.DataFrame,
    num_genes: int,
    max_iters: int = 2000,
) -> List[str]:
    """
    Identify genes on the Pareto frontier of a volcano plot with respect to MAE
    and time correlation.

    Args:
        volcano_data (pd.DataFrame): DataFrame containing MAE and time correlation
        num_genes (int): Number of genes to return.
        max_iters (int, optional): Maximum number of iterations. Defaults to 2000.

    Returns:
        List[str]: List of gene indices from volcano_data.
    """
    volcano_data = volcano_data.loc[
        ~volcano_data.index.str.contains(("^Rpl|^Rps"), case=False)
    ]
    pareto_frontier = pd.DataFrame()

    if len(volcano_data) < num_genes:
        logger.info(
            f"\nWarning: Not enough genes in the input data.\n"
            f"Only {len(volcano_data)} genes were found,\n"
            f"but {num_genes} were requested.\n"
            "Attempting to return as many as possible.\n"
        )

    pareto_frontier_rows = []

    counter = 0

    while (
        len(pareto_frontier_rows) < num_genes
        and len(volcano_data) > 0
        and counter < max_iters
    ):
        counter += 1
        sorted_data = volcano_data.sort_values(
            by=["mean_mae", "time_correlation"], ascending=[False, False]
        )
        pareto_frontier_current = [sorted_data.iloc[[0]]]

        for i in range(1, len(sorted_data)):
            if (
                sorted_data["time_correlation"].iloc[i]
                >= pareto_frontier_current[-1]["time_correlation"].iloc[0]
            ):
                pareto_frontier_current.append(sorted_data.iloc[[i]])
        pareto_frontier_indices = [
            pf.index[0] for pf in pareto_frontier_current
        ]
        volcano_data = volcano_data.drop(pareto_frontier_indices)
        pareto_frontier_rows.extend(pareto_frontier_current)

        logger.info(
            f"\nFound {len(pareto_frontier_current)} genes on the current Pareto frontier:\n\n"
            f"  {pareto_frontier_indices}\n\n"
            f"Genes identified thus far: {len(pareto_frontier_rows)}.\n"
            f"Number of iterations: {counter}.\n"
            f"Number of genes remaining: {len(volcano_data)}.\n\n"
        )
        if counter >= max_iters:
            logger.warning(
                f"\nWarning: Maximum number of iterations reached ({max_iters}).\n"
                f"Returning {len(pareto_frontier)} genes found so far.\n\n"
            )

    pareto_frontier = pd.concat(
        pareto_frontier_rows[:num_genes]
    ).drop_duplicates()
    pareto_frontier = pareto_frontier.sort_values(
        by="time_correlation", ascending=False
    )

    gene_indices = pareto_frontier.index.tolist()
    logger.info(
        f"\nFound {len(pareto_frontier)} genes on the Pareto frontier for {num_genes} requested:\n\n"
        f"  {gene_indices}\n\n"
    )
    return gene_indices



from os import PathLike
from typing import Dict

import numpy as np
from beartype import beartype
from matplotlib import pyplot as plt
from matplotlib.figure import FigureBase

from pyrovelocity.logging import configure_logging


logger = configure_logging(__name__)

__all__ = ["plot_t0_selection"]


@beartype
def plot_t0_selection(
    posterior_samples: Dict[str, np.ndarray],
    t0_selection_plot: PathLike | str,
) -> FigureBase:
    """
    Plot the t0, switching time, and cell time posterior samples.

    Args:
        posterior_samples (Dict[str, np.ndarray]): Dictionary of posterior samples.
        t0_selection_plot (PathLike | str): Path to save the plot.

    Returns:
        FigureBase: Matplotlib figure object.
    """
    fig, ax = plt.subplots(5, 6)
    fig.set_size_inches(26, 24)
    ax = ax.flatten()

    posterior_samples_keys_to_check = ["t0", "switching", "cell_time"]

    num_samples_list = [
        posterior_samples[key].shape[0]
        for key in posterior_samples_keys_to_check
    ]
    assert (
        len(set(num_samples_list)) == 1
    ), f"The number of samples is not equal across keys: {posterior_samples_keys_to_check}"

    num_samples = num_samples_list[0]

    for sample in range(num_samples):
        t0_sample = posterior_samples["t0"][sample]
        switching_sample = posterior_samples["switching"][sample]
        cell_time_sample = posterior_samples["cell_time"][sample]
        ax[sample].scatter(
            t0_sample.flatten(),
            2 * np.ones(t0_sample.shape[-1]),
            s=1,
            c="red",
            label="t0",
        )
        ax[sample].scatter(
            switching_sample.flatten(),
            3 * np.ones(t0_sample.shape[-1]),
            s=1,
            c="purple",
            label="switching",
        )
        ax[sample].scatter(
            cell_time_sample.flatten(),
            np.ones(cell_time_sample.shape[0]),
            s=1,
            c="blue",
            label="shared time",
        )
        ax[sample].set_ylim(-0.5, 4)
        if sample == 28:
            ax[sample].legend(loc="best", bbox_to_anchor=(0.5, 0.0, 0.5, 0.5))
    ax[-1].hist(t0_sample.flatten(), bins=200, color="red", alpha=0.3)
    ax[-1].hist(cell_time_sample.flatten(), bins=500, color="blue", alpha=0.3)

    fig.savefig(
        t0_selection_plot,
        facecolor=fig.get_facecolor(),
        bbox_inches="tight",
        edgecolor="none",
        dpi=300,
    )

    return fig


# from typing import Dict
# from typing import List

# import anndata
# import matplotlib
# import matplotlib.pyplot as plt
# import numpy as np
# import pandas as pd
# import pyro
# import scvelo as scv
# import seaborn as sns
# import torch
# from adjustText import adjust_text
# from anndata import AnnData
# from matplotlib import cm
# from matplotlib import gridspec
# from matplotlib.colors import Normalize
# from matplotlib.figure import Figure
# from matplotlib.ticker import MaxNLocator
# from mpl_toolkits.axes_grid1 import make_axes_locatable
# from numpy import ndarray
# from scipy.sparse import issparse
# from scipy.stats import spearmanr
# from scvelo.plotting.velocity_embedding_grid import default_arrow

# from pyrovelocity.analyze import compute_mean_vector_field
# from pyrovelocity.analyze import compute_volcano_data
# from pyrovelocity.models import mrna_dynamics


# def plot_evaluate_dynamic_orig(adata, gene="Cpe", velocity=None, ax=None):
#     # compute dynamics
#     alpha, beta, gamma, scaling, t_ = (
#         torch.tensor(adata.var.loc[gene, "fit_alpha"]),
#         torch.tensor(adata.var.loc[gene, "fit_beta"]),
#         torch.tensor(adata.var.loc[gene, "fit_gamma"]),
#         torch.tensor(adata.var.loc[gene, "fit_scaling"]),
#         torch.tensor(adata.var.loc[gene, "fit_t_"]),
#     )

#     beta_scale = beta * scaling

#     u0, s0 = adata.var.loc[gene, "fit_u0"], adata.var.loc[gene, "fit_s0"]
#     # t = torch.tensor(adata[:, gene].layers['fit_t'][:, 0]).sort()[0]
#     t = torch.tensor(adata[:, gene].layers["fit_t"][:, 0])
#     u_inf, s_inf = mrna_dynamics(t_, u0, s0, alpha, beta_scale, gamma)

#     state = (t < t_).int()
#     tau = t * state + (t - t_) * (1 - state)
#     u0_vec = u0 * state + u_inf * (1 - state)
#     s0_vec = s0 * state + s_inf * (1 - state)

#     alpha_ = 0
#     alpha_vec = alpha * state + alpha_ * (1 - state)

#     ut, st = mrna_dynamics(tau, u0_vec, s0_vec, alpha_vec, beta_scale, gamma)
#     ut = ut * scaling + u0
#     st = st + s0

#     xnew = torch.linspace(torch.min(st), torch.max(st))
#     ynew = gamma / beta * (xnew - torch.min(xnew)) + torch.min(ut)

#     if ax is None:
#         fig, ax = plt.subplots()
#         scv.pl.scatter(adata, gene, color=["clusters"], ax=ax, show=False)
#     else:
#         ax.scatter(
#             st.detach().numpy(),
#             ut.detach().numpy(),
#             linestyle="-",
#             linewidth=5,
#             alpha=0.3,
#         )
#         ax.plot(
#             xnew.detach().numpy(),
#             ynew.detach().numpy(),
#             color="b",
#             linestyle="--",
#             linewidth=5,
#         )

#     if velocity is None:
#         print(
#             "scvelo %s mse loss:" % gene,
#             mse_loss_sum(
#                 ut,
#                 st,
#                 adata[:, gene].layers["Mu"].toarray()[:, 0],
#                 adata[:, gene].layers["Ms"].toarray()[:, 0],
#             ),
#         )
#     else:
#         print(
#             "scvelo %s mse loss:" % gene,
#             mse_loss_sum(
#                 ut[velocity.weight],
#                 st[velocity.weight],
#                 adata[:, gene].layers["Mu"].toarray()[velocity.weight, 0],
#                 adata[:, gene].layers["Ms"].toarray()[velocity.weight, 0],
#             ),
#         )
#     return alpha_vec, ut, st, xnew, ynew


# def plot_dynamic_pyro(
#     adata,
#     gene,
#     losses,
#     summary,
#     velocity,
#     fix_param_list,
#     alpha,
#     beta,
#     gamma,
#     scale,
#     t_,
#     t,
# ):
#     alpha = (
#         torch.tensor(alpha)
#         if fix_param_list[0] == 1
#         else pyro.param("AutoDelta.alpha_sample")
#     )
#     beta = (
#         torch.tensor(beta)
#         if fix_param_list[1] == 1
#         else pyro.param("AutoDelta.beta_sample")
#     )
#     gamma = (
#         torch.tensor(gamma)
#         if fix_param_list[2] == 1
#         else pyro.param("AutoDelta.gamma_sample")
#     )
#     scale = (
#         torch.tensor(scale)
#         if fix_param_list[3] == 1
#         else pyro.param("AutoDelta.scale_sample")
#     )
#     t_ = (
#         torch.tensor(t_)
#         if fix_param_list[4] == 1
#         else pyro.param("AutoDelta.switching_sample")
#     )

#     if fix_param_list[5] == 0:
#         t = pyro.param("AutoDelta.latent_time")
#     else:
#         t = torch.tensor(t)
#     fig, ax = plt.subplots(1, 3)
#     fig.set_size_inches(16, 4)
#     ax[1].scatter(
#         adata[:, gene].layers["fit_t"].toarray()[velocity.weight, 0],
#         t.data.cpu().numpy(),
#     )

#     t = (t.sort()[0].max() + 1).int()
#     t = torch.linspace(0.0, t, 500)

#     beta_scale = beta * scale
#     u0, s0 = torch.tensor(0.0), torch.tensor(0.0)
#     u_inf, s_inf = mrna_dynamics(t_, u0, s0, alpha, beta_scale, gamma)

#     state = (t < t_).int()
#     tau = t * state + (t - t_) * (1 - state)
#     u0_vec = u0 * state + u_inf * (1 - state)
#     s0_vec = s0 * state + s_inf * (1 - state)

#     alpha_ = 0.0
#     alpha_vec = alpha * state + alpha_ * (1 - state)
#     ut, st = mrna_dynamics(tau, u0_vec, s0_vec, alpha_vec, beta_scale, gamma)
#     ut = ut * scale + u0
#     st = st + s0
#     xnew = torch.linspace(
#         torch.tensor(st.min().detach().numpy()),
#         torch.tensor(st.max().detach().numpy()),
#         50,
#     )
#     ynew = gamma / beta * (xnew - torch.min(xnew)) + torch.min(ut)

#     ax[0].plot(losses)
#     ax[0].set_yscale("log")
#     ax[0].set_title("ELBO")
#     ax[0].set_xlabel("step")
#     ax[0].set_ylabel("loss")

#     scv.pl.scatter(adata, gene, color=["clusters"], ax=ax[2], show=False)
#     ax[2].scatter(
#         summary["x_obs"]["mean"][:, 1],
#         summary["x_obs"]["mean"][:, 0],
#         alpha=0.5,
#         s=5,
#         color="r",
#     )
#     # ax[2].scatter(summary['x_obs']['5%'][:, 1], summary['x_obs']['5%'][:, 0], alpha=0.2, color='r')
#     # ax[2].scatter(summary['x_obs']['95%'][:, 1], summary['x_obs']['95%'][:, 0], alpha=0.2, color='r')
#     ax[2].plot(
#         st.detach().numpy(),
#         ut.detach().numpy(),
#         linestyle="-",
#         linewidth=5,
#         color="g",
#     )
#     ax[2].plot(
#         xnew.detach().numpy(),
#         ynew.detach().numpy(),
#         color="g",
#         linestyle="--",
#         linewidth=5,
#     )
#     # ax[2].set_ylim(0, 3)
#     # ax[2].set_xlim(0, 16)
#     plot_evaluate_dynamic_orig(adata, gene, velocity, ax[2])

#     print(
#         "pyro model mse loss",
#         mse_loss_sum(
#             summary["x_obs"]["mean"][:, 0],
#             summary["x_obs"]["mean"][:, 1],
#             velocity.x[:, 0],
#             velocity.x[:, 1],
#         ),
#     )
#     return alpha_vec, ut, st, xnew, ynew


# def plot_multigenes_dynamical(
#     summary,
#     alpha,
#     beta,
#     gamma,
#     t_,
#     t,
#     adata,
#     gene="Cpe",
#     scale=None,
#     ax=None,
#     raw=False,
# ):
#     pass

#     # softplus operation as pyro
#     # https://stackoverflow.com/questions/44230635/avoid-overflow-with-softplus-function-in-python
#     # t_ = torch.log(1+torch.exp(-np.abs(t_))) + torch.maximum(t_,
#     #     torch.zeros(t_.shape))

#     t = t.sort()[0].max().int()
#     t = torch.linspace(0.0, t, 500)
#     u0, s0 = torch.tensor(0.0), torch.tensor(0.0)
#     # u0, s0 = pyro.param("u0"), pyro.param("s0")
#     u_inf, s_inf = mrna_dynamics(t_, u0, s0, alpha, beta, gamma)
#     state = (t < t_).int()
#     tau = t * state + (t - t_) * (1 - state)

#     # tau = torch.log(1+torch.exp(-np.abs(tau))) + torch.maximum(tau,
#     #    torch.zeros(tau.shape))

#     u0_vec = u0 * state + u_inf * (1 - state)
#     s0_vec = s0 * state + s_inf * (1 - state)
#     alpha_ = 0.0
#     alpha_vec = alpha * state + alpha_ * (1 - state)
#     ut, st = mrna_dynamics(tau, u0_vec, s0_vec, alpha_vec, beta, gamma)
#     if scale is None:
#         ut = ut + u0
#     else:
#         ut = ut * scale + u0
#     st = st + s0
#     xnew = torch.linspace(
#         torch.tensor(st.min().detach().numpy()),
#         torch.tensor(st.max().detach().numpy()),
#         50,
#     )
#     if scale is not None:
#         ynew = (gamma / beta * (xnew - torch.min(xnew))) * scale + torch.min(
#             ut * scale + u0
#         )
#     else:
#         ynew = gamma / beta * (xnew - torch.min(xnew)) + torch.min(ut)

#     if ax is None:
#         fig, ax = plt.subplots()
#     try:
#         if raw:
#             scv.pl.scatter(
#                 adata,
#                 gene,
#                 x="spliced",
#                 y="unspliced",
#                 color=["clusters"],
#                 ax=ax,
#                 show=False,
#             )
#         else:
#             scv.pl.scatter(adata, gene, color=["clusters"], ax=ax, show=False)
#     except:
#         if raw:
#             scv.pl.scatter(
#                 adata,
#                 gene,
#                 x="spliced",
#                 y="unspliced",
#                 color=["Clusters"],
#                 ax=ax,
#                 show=False,
#             )
#         else:
#             scv.pl.scatter(adata, gene, color=["Clusters"], ax=ax, show=False)
#     ax.plot(
#         st.detach().numpy(),
#         ut.detach().numpy(),
#         linestyle="-",
#         linewidth=2.5,
#         color="red",
#         label="Pyro-Velocity",
#     )
#     ax.plot(
#         xnew.detach().numpy(),
#         ynew.detach().numpy(),
#         color="red",
#         linestyle="--",
#         linewidth=2.5,
#         alpha=0.4,
#     )
#     if summary is not None:
#         ax.scatter(
#             summary["x_obs"]["mean"][:, 1],
#             summary["x_obs"]["mean"][:, 0],
#             alpha=0.5,
#             color="red",
#         )
# def denoised_umap(posterior_samples, adata, cell_state="state_info"):
#     pass

#     import sklearn
#     import umap
#     from sklearn.pipeline import Pipeline

#     projection = [
#         ("PCA", sklearn.decomposition.PCA(random_state=99, n_components=50)),
#         ("UMAP", umap.UMAP(random_state=99, n_components=2)),
#     ]
#     pipelines = Pipeline(projection)
#     fig, ax = plt.subplots(2, 2)
#     fig.set_size_inches(9, 9)
#     expression = [posterior_samples["st"].mean(0)]
#     pipelines.fit(expression[0])
#     umap_orig = pipelines.transform(expression[0])
#     adata.obsm["X_umap1"] = umap_orig
#     scv.pl.scatter(adata, basis="umap1", ax=ax[0][0], show=False)

#     joint_pcs = pipelines.steps[0][1].transform(expression[0])
#     adata.obsm["X_pyropca"] = joint_pcs
#     scv.pp.neighbors(adata, use_rep="pyropca")
#     adata.layers["spliced_pyro"] = posterior_samples["st"].mean(0)
#     if "u_scale" in posterior_samples:
#         adata.layers["velocity_pyro"] = (
#             posterior_samples["ut"]
#             * posterior_samples["beta"]
#             / (posterior_samples["u_scale"] / posterior_samples["s_scale"])
#             - posterior_samples["st"] * posterior_samples["gamma"]
#         ).mean(0)
#     else:
#         adata.layers["velocity_pyro"] = (
#             posterior_samples["ut"] * posterior_samples["beta"]
#             - posterior_samples["st"] * posterior_samples["gamma"]
#         ).mean(0)
#     scv.tl.velocity_graph(adata, vkey="velocity_pyro", xkey="spliced_pyro")
#     scv.tl.velocity_embedding(adata, vkey="velocity_pyro", basis="umap1")
#     scv.pl.velocity_embedding_grid(
#         adata,
#         basis="umap1",
#         vkey="velocity_pyro",
#         density=0.5,
#         scale=0.25,
#         arrow_size=3,
#         color=cell_state,
#         ax=ax[0][1],
#         show=False,
#     )
#     adata.obsm["X_umap1"] = umap_orig

#     expression = [
#         np.hstack(
#             [posterior_samples["st"].mean(0), posterior_samples["ut"].mean(0)]
#         )
#     ]
#     pipelines.fit(expression[0])
#     umap_orig = pipelines.transform(expression[0])
#     adata.obsm["X_umap2"] = umap_orig
#     scv.pl.scatter(adata, basis="umap2", ax=ax[1][0], show=False)
#     joint_pcs = pipelines.steps[0][1].transform(expression[0])
#     adata.obsm["X_pyropca"] = joint_pcs
#     scv.pp.neighbors(adata, use_rep="pyropca")
#     adata.layers["spliced_pyro"] = posterior_samples["st"].mean(0)
#     if "u_scale" in posterior_samples:
#         adata.layers["velocity_pyro"] = (
#             posterior_samples["ut"]
#             * posterior_samples["beta"]
#             / (posterior_samples["u_scale"] / posterior_samples["s_scale"])
#             - posterior_samples["st"] * posterior_samples["gamma"]
#         ).mean(0)
#     else:
#         adata.layers["velocity_pyro"] = (
#             posterior_samples["ut"] * posterior_samples["beta"]
#             - posterior_samples["st"] * posterior_samples["gamma"]
#         ).mean(0)
#     scv.tl.velocity_graph(adata, vkey="velocity_pyro", xkey="spliced_pyro")
#     scv.tl.velocity_embedding(adata, vkey="velocity_pyro", basis="umap1")
#     scv.pl.velocity_embedding_grid(
#         adata,
#         basis="umap2",
#         vkey="velocity_pyro",
#         color=cell_state,
#         density=0.5,
#         scale=0.25,
#         arrow_size=3,
#         show=False,
#         ax=ax[1][1],
#     )

# def mse_loss_sum(u_model, s_model, u_data, s_data):
#     """
#     Computes the mean squared error loss sum between the model and data.

#     Args:
#         u_model (torch.Tensor): Predicted values of u from the model.
#         s_model (torch.Tensor): Predicted values of s from the model.
#         u_data (torch.Tensor): True values of u from the data.
#         s_data (torch.Tensor): True values of s from the data.

#     Returns:
#         torch.Tensor: Mean squared error loss sum.

#     Examples:
#         >>> import torch
#         >>> u_model = torch.tensor([0.5, 0.6])
#         >>> s_model = torch.tensor([0.7, 0.8])
#         >>> u_data = torch.tensor([0.4, 0.5])
#         >>> s_data = torch.tensor([0.6, 0.7])
#         >>> mse_loss_sum(u_model, s_model, u_data, s_data)
#         tensor(0.0200)
#     """
#     return ((u_model - u_data) ** 2 + (s_model - s_data) ** 2).mean(0)



import dataclasses
import importlib
import inspect
import logging
import os
import pkgutil
import queue
import secrets
import sys
import threading
import time
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import is_dataclass
from datetime import timedelta
from textwrap import dedent
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import get_args
from typing import get_origin

from dataclasses_json import dataclass_json
from dulwich.repo import NotGitRepository
from dulwich.repo import Repo
from flytekit import WorkflowExecutionPhase
from flytekit.core.base_task import PythonTask
from flytekit.core.workflow import WorkflowBase
from flytekit.exceptions.system import FlyteSystemException
from flytekit.exceptions.user import FlyteTimeout
from flytekit.remote import FlyteRemote
from flytekit.remote.executions import FlyteWorkflowExecution
from hydra.conf import HelpConf
from hydra.conf import HydraConf
from hydra.conf import JobConf
from hydra_zen import ZenStore
from hydra_zen import builds
from hydra_zen import make_custom_builds_fn

from pyrovelocity.logging import configure_logging


logger = configure_logging(__name__)


@dataclass_json
@dataclass
class EntityConfig:
    inputs: Dict[str, Any]
    module_name: str = "main_workflow"
    entity_name: str = "training_workflow"
    entity_type: str = "WorkflowBase"


# @dataclass_json
# @dataclass
# class EntityConfigs:
#     entities: Dict[str, Any] = MISSING


fbuilds = make_custom_builds_fn(populate_full_signature=True)


def generate_entity_configs(
    parent_module_path: str, entity_store: ZenStore, logger: logging.Logger
) -> None:
    """
    Generates and stores configurations for entities found in a specified
    module.

    This function iterates over all submodules in the parent module defined by
    `parent_module_path`. For each submodule, it looks for entities that are
    instances of either `WorkflowBase` or `PythonTask` and generates
    configurations for them using `EntityConfig`. These configurations are then
    stored in `entity_store`.

    Args:
        parent_module_path (str): The import path of the parent module to search
        for entities.
        entity_store (ZenStore): The store where generated entity configurations
        will be kept.
        logger (logging.Logger): Logger for debugging and logging information.

    Returns:
        None: The function does not return anything. It populates the
        `entity_store` with configurations for each entity of type EntityTypes.
    """
    parent_module = importlib.import_module(parent_module_path)
    EntityTypes = (WorkflowBase, PythonTask)

    # iterate over submodules in the parent module
    for submodule_info in pkgutil.iter_modules(
        parent_module.__path__, parent_module.__name__ + "."
    ):
        # import the submodule
        submodule = importlib.import_module(submodule_info.name)
        logger.debug(f"Checking submodule: {submodule_info.name}")

        # import entities that are instances of EntityTypes
        entities = inspect.getmembers(
            submodule,
            # TODO: validate that PythonTasks function as expected
            lambda member: isinstance(member, EntityTypes),
            # lambda member: isinstance(member, WorkflowBase),
        )

        for entity_name, entity in entities:
            logger.debug(f"Found entity: {entity_name}")

            # construct an instance (or a configuration) of the entity
            module_name = submodule_info.name.split(".")[-1]
            entity_inputs = generate_entity_inputs(entity)
            entity_instance = fbuilds(
                EntityConfig,
                inputs=builds(dict, entity_inputs, hydra_convert="all"),
                module_name=module_name,
                entity_name=entity_name,
                entity_type=type(entity).__name__,
            )

            # store the entity instance in the entity_store
            composed_name = module_name + "_" + entity_name
            entity_store(entity_instance, name=composed_name)
            logger.debug(f"Stored entity: {composed_name} in entity_store")


def generate_entity_inputs(
    entity: Union[WorkflowBase, PythonTask],
) -> Dict[str, Any]:
    """
    Generates a dictionary of inputs for a given entity.

    This function inspects the signature of the provided `entity`, which can be
    either a `WorkflowBase` or a `PythonTask`. For each parameter in the
    signature, it determines the type and default value (if any). If the type is
    a built-in type, it directly uses the default value. For custom types, it
    dynamically imports and constructs a configuration object using `fbuilds`.

    Args:
        entity (Union[WorkflowBase, PythonTask]): The entity for which to
        generate input configurations.

    Returns:
        Dict[str, Any]: A dictionary with keys that are the names of the entity
        inputs and values hydra-zen configurations that will build their
        respective default values.
    """
    inputs = {}

    # TODO: add support for dataclasses with fields of list[DataClassName]
    # def build_config_for_list(list_items, item_type):
    #     configs = []
    #     for item in list_items:
    #         if is_dataclass(item):
    #             item_config = {
    #                 field.name: build_config(
    #                     field.type, getattr(item, field.name)
    #                 )
    #                 for field in dataclasses.fields(item)
    #             }
    #             configs.append(fbuilds(item_type, **item_config))
    #         else:
    #             configs.append(item)
    #     return configs

    def build_config(param_type, value):
        # if get_origin(param_type) is list:
        #     item_type = get_args(param_type)[0]
        #     return build_config_for_list(value, item_type)
        if is_dataclass(value):
            field_configs = {
                field.name: build_config(field.type, getattr(value, field.name))
                for field in dataclasses.fields(value)
            }
            return fbuilds(param_type, **field_configs)
        else:
            return value

    for name, param in inspect.signature(entity).parameters.items():
        param_type = param.annotation
        default = param.default
        logger.debug(
            f"Found parameter: {name}, Type: {param_type}, Default: {default}"
        )

        # if isinstance(default, list):
        #     inputs[name] = build_config(param_type, default)
        if is_dataclass(default):
            inputs[name] = build_config(param_type, default)
        elif (
            isinstance(param_type, type) and param_type.__module__ == "builtins"
        ):
            inputs[name] = default
        else:
            type_module = importlib.import_module(param_type.__module__)
            custom_type = getattr(type_module, param_type.__name__)
            inputs[name] = fbuilds(custom_type)

    return inputs


def random_alphanumeric_suffix(input_string: str = "", length: int = 3) -> str:
    return input_string.join(
        secrets.choice("abcdefghijklmnopqrstuvwxyz0123456789")
        for _ in range(length)
    )


def check_required_env_vars(
    required_vars: List[str], logger: logging.Logger
) -> bool:
    """
    Checks required environment variables for workflow configuration.
    """

    missing_vars = [var for var in required_vars if os.environ.get(var) is None]
    if missing_vars:
        missing_vars_str = ", ".join(missing_vars)
        logger.error(
            f"Missing required environment variables: {missing_vars_str}"
        )
        return False
    return True


def git_info_to_workflow_version(
    logger: logging.Logger,
) -> Tuple[str, str, str]:
    """
    Retrieves git information for workflow versioning using Dulwich.

    This function extracts repository name, current branch name, and short SHA.
    It handles the case where the Git repository is in a detached HEAD state,
    common in CI environments like GitHub Actions for pull requests.

    Args:
        logger (logging.Logger): Logger object for logging messages.

    Returns:
        Tuple[str, str, str]: A tuple containing the repository name,
                               branch name, and short SHA commit.

    Raises:
        ValueError: If unable to extract source commit SHA from commit message.

    Example:
        >>> import logging
        >>> logger = logging.getLogger()
        >>> # assuming this test is run in a git repository
        >>> repo_name, branch, short_sha = git_info_to_workflow_version(logger)
        >>> print(repo_name, branch, short_sha)
        >>> print(isinstance(repo_name, str), isinstance(branch, str), isinstance(short_sha, str))
        True True True
    """
    try:
        try:
            repo = Repo(".")

            config_stack = repo.get_config_stack()
            remote_url = config_stack.get(
                (b"remote", b"origin"), b"url"
            ).decode()
            repo_name = os.path.basename(remote_url.rstrip("/")).rstrip(".git")

            head_commit = repo.head()
            branches = {
                name.decode(): sha
                for name, sha in repo.refs.as_dict(b"refs/heads").items()
            }
            logger.debug(f"Found branches:\n{branches}")
            branch_name = None
            short_sha = None
            for name, sha in branches.items():
                if sha == repo.head():
                    branch_name = name
                    short_sha = head_commit.decode("utf-8")[:7]
                    break

            if branch_name is None:
                logger.warning(
                    "Repository may be in detached HEAD state or missing ref."
                    "You may try running `git fetch +refs/heads/*:refs/remotes/origin/*`."
                    "Falling back to GIT_REPO_NAME, GIT_REF, GIT_SHA_SHORT env vars"
                    "or noref defaults: CWD, nobranch, 0000000."
                )
                repo_name = os.environ.get(
                    "GIT_REPO_NAME", os.path.basename(os.getcwd())
                )
                branch_name = os.environ.get("GIT_REF", "nobranch")
                short_sha = os.environ.get("GIT_SHA_SHORT", "0000000")

        except NotGitRepository:
            no_git_repo_fallback_message = (
                "May not be in a Git repository."
                "Falling back to GIT_REPO_NAME, GIT_REF, GIT_SHA_SHORT env vars"
                "or CWD, nobranch, 0000000 defaults."
            )
            logger.warning(no_git_repo_fallback_message)
            repo_name = os.environ.get(
                "GIT_REPO_NAME", os.path.basename(os.getcwd())
            )
            branch_name = os.environ.get("GIT_REF", "nobranch")
            short_sha = os.environ.get("GIT_SHA_SHORT", "0000000")

        repo_name = repo_name.lower()
        branch_name = branch_name.lower()
        short_sha = short_sha.lower()

        return repo_name, branch_name, short_sha

    except Exception as e:
        logger.error(f"Error obtaining git information: {e}")
        return "norepo", "nobranch", "0000000"


def generate_hydra_config() -> HydraConf:
    return HydraConf(
        defaults=[
            {"output": "default"},
            {"launcher": "basic"},  # joblib
            {"sweeper": "basic"},
            {"help": "default"},
            {"hydra_help": "default"},
            {"hydra_logging": "none"},  # default
            {"job_logging": "none"},  # default
            {"callbacks": None},
            {"env": "default"},
        ],
        help=HelpConf(
            header=dedent(
                """
                This is the ${hydra.help.app_name} help accessible via `${hydra.help.app_name} -h`.

                Use `${hydra.help.app_name} -c job` to view the ${hydra.help.app_name} configuration alone.
                See the end of this help page for instructions on how to install shell tab completion for
                configuration overrides.

                ${hydra.help.app_name} is the CLI of a template designed to illustrate the integration of:

                  * hydra-zen (https://mit-ll-responsible-ai.github.io/hydra-zen/),
                  * hydra (https://hydra.cc/), and
                  * omegaconf (https://omegaconf.readthedocs.io/),

                which provide configuration management, with

                  * flyte(kit) (https://flyte.org/),

                which manages the registration and execution of Flyte workflows.
                ${hydra.help.app_name} can be adapted as an auxiliary component of any python package,
                enhancing its capabilities in managing complex workflow configuration
                and execution.

                Running `${hydra.help.app_name} -c job` displays the current configuration of ${hydra.help.app_name}.
                This reflects what will be executed if `pyrovelocity` is run without arguments.

                Use `${hydra.help.app_name} -c hydra` to view the associated hydra configuration.

                """
            ),
            footer=dedent(
                """
                You can test CLI configuration overrides after `-c job`, e.g.:

                  * `${hydra.help.app_name} -c job execution_context=prod`
                  * `${hydra.help.app_name} -c job entity_config==example_wf`
                  # The inputs arguments in the following example must correspond
                  # to those supported by the specified entity_config.
                  * `${hydra.help.app_name} -c job \\
                       entity_config=main_workflow_training_workflow \\
                       entity_config.inputs._args_.0.logistic_regression.max_iter=1200`
                  # The following two examples are only meant for task testing.
                  # This example only works in the local_shell execution context.
                  # See: https://github.com/flyteorg/flyte/issues/4275
                  #      https://github.com/flyteorg/flyte/issues/1312
                  * `${hydra.help.app_name} -c job \\
                       execution_context=local_shell \\
                       entity_config=main_workflow_process_data \\
                       entity_config.inputs._args_=[]"`
                  # This will fail without specifying the inputs as it does not
                  # automatically instantiate the default arguments of the task.
                  * `${hydra.help.app_name} -c job \\
                       execution_context=local_cluster_dev \\
                       entity_config=main_workflow_process_data \\
                       entity_config.inputs._args_.0.data.data="[[12.0, 0],[13.0, 1],[9.5, 2]]" \\
                       entity_config.inputs._args_.0.data.columns="[ash, target]"`


                This will generate `== Config ==` above resolved in context of the command line overrides.
                Removing the `-c job` flag will execute the workflow with the specified configuration.
                The resolved configuration will be stored in the `outputs` or `multirun` directories.

                Use `${hydra.help.app_name} --hydra-help` to view the hydra
                help. This contains, for example, the commands to install shell
                tab completion. For example in bash or zsh, if the active
                configuration has path
                `entity_config.inputs._args_.0.logistic_regression`
                representing the parameters of a
                sklearn.linear_model.LogisticRegression instance:

                > eval "$$(pyrovelocity -sc install=bash)"
                > pyrovelocity entity_config.inputs._args_.0.logistic_regression.[TAB]
                entity_config.inputs._args_.0.logistic_regression.C=
                entity_config.inputs._args_.0.logistic_regression._target_=
                entity_config.inputs._args_.0.logistic_regression.class_weight=
                entity_config.inputs._args_.0.logistic_regression.dual=
                ..."""
            ),
            template=dedent(
                """
                ${hydra.help.header}
                == Configuration groups ==
                First override default group values (group=option)

                $APP_CONFIG_GROUPS

                == Config ==
                Then override any element in the config (foo.bar=value)
                that is not set exclusively by an environment variable [see doc(strings)]

                $CONFIG
                ${hydra.help.footer}
                """
            ),
        ),
        job=JobConf(name="pyrovelocity"),
    )


def get_user_input(input_queue):
    """
    Gets user input and puts it in the queue.
    """
    user_input = input("Terminate workflow execution? (y/N after 1 min.): ")
    input_queue.put(user_input)


def wait_for_workflow_completion(
    execution: FlyteWorkflowExecution,
    remote: FlyteRemote,
    logger: logging.Logger,
) -> None:
    """
    Waits for the execution to complete, checking status at regular intervals.
    """
    timeout_duration = timedelta(seconds=3.0)
    synced_execution = None
    try:
        while True:
            try:
                completed_execution = remote.wait(
                    execution, timeout=timeout_duration
                )
                logger.info(f"Execution completed:\n\n{completed_execution}\n")
                if completed_execution.error is None:
                    break
                else:
                    logger.error(
                        f"Execution failed with error:\n\n{completed_execution.error}\n"
                    )
                    sys.exit(1)
            except FlyteTimeout:
                synced_execution = remote.sync(execution)
                logger.info(f"Current status:\n\n{synced_execution}\n")
                time.sleep(timeout_duration.total_seconds())
    except KeyboardInterrupt:
        if synced_execution is not None:
            logger.info(f"Status at KeyboardInterrupt:\n\n{synced_execution}\n")
        else:
            logger.info(
                "KeyboardInterrupt caught before execution status sync."
            )

        input_queue = queue.Queue()
        input_thread = threading.Thread(
            target=get_user_input, args=(input_queue,)
        )
        input_thread.daemon = True
        input_thread.start()

        try:
            response = input_queue.get(timeout=60)
            response = response.strip().lower()
        except queue.Empty:
            response = "n"

        synced_execution = remote.sync(execution)
        if synced_execution.closure.phase in [WorkflowExecutionPhase.RUNNING]:
            try:
                if response in ["y", "yes"]:
                    remote.terminate(
                        execution, "KeyboardInterrupt confirmed termination"
                    )
                    logger.info("Workflow execution terminated.")
                else:
                    logger.warning(
                        f"\nExiting script without terminating workflow execution:\n\n{execution}\n"
                    )
            except FlyteSystemException as e:
                logger.error(
                    f"Error while trying to terminate the execution: {e}"
                )
        else:
            logger.info(
                f"Workflow execution already in terminal state: {synced_execution.closure.phase}"
            )

        sys.exit()


# -----------
# DEPRECATED
# -----------


def generate_workflow_inputs(
    workflow_import_path: str = "pyrovelocity.workflows.main_workflow",
    workflow_name: str = "training_workflow",
) -> Dict[str, Any]:
    """
    Deprecated in favor of `generate_entity_inputs`.
    """
    module = importlib.import_module(workflow_import_path)
    workflow = getattr(module, workflow_name)

    if not callable(workflow):
        value_error_message = f"Workflow '{workflow_name}' is not callable"
        raise ValueError(value_error_message)

    inputs = {}

    for name, param in inspect.signature(workflow).parameters.items():
        param_type = param.annotation
        default = param.default

        # check if the type is a built-in type (like int, str, etc.)
        if isinstance(param_type, type) and param_type.__module__ == "builtins":
            inputs[name] = default
        else:
            # dynamically import the type if it's not a built-in type
            type_module = importlib.import_module(param_type.__module__)
            custom_type = getattr(type_module, param_type.__name__)

            inputs[name] = builds(custom_type)

    return inputs


# def git_info_to_workflow_version(
#     logger: logging.Logger,
# ) -> Tuple[str, str, str]:
#     """
#     Retrieves git information for workflow versioning using plumbum.

#     This function extracts repository name, current branch name, and short SHA.
#     It handles the case where the Git repository is in a detached HEAD state,
#     common in CI environments like GitHub Actions for pull requests.

#     Args:
#         logger (logging.Logger): Logger object for logging messages.

#     Returns:
#         Tuple[str, str, str]: A tuple containing the repository name,
#                                branch name, and short SHA commit.

#     Raises:
#         ValueError: If unable to extract source commit SHA from commit message.
#         ProcessExecutionError: If a git command fails.

#     Example:
#         >>> import logging
#         >>> logger = logging.getLogger()
#         >>> # Assuming this test is run in a Git repository
#         >>> repo_name, branch, short_sha = git_info_to_workflow_version(logger)
#         >>> print(isinstance(repo_name, str), isinstance(branch, str), isinstance(short_sha, str))
#         True True True
#     """
#     try:
#         git = plumbum.local["git"]
#         git_branch = git("rev-parse", "--abbrev-ref", "HEAD").strip()

#         if git_branch.lower() == "head":
#             git("fetch", "origin", "+refs/heads/*:refs/remotes/origin/*")

#             commit_message = git("log", "-1", "--pretty=%B")
#             match = re.search(r"Merge ([0-9a-f]{40}) into", commit_message)
#             if match:
#                 source_commit_sha = match.group(1)

#                 git_branch_list = (
#                     git("branch", "-r", "--contains", source_commit_sha)
#                     .strip()
#                     .split("\n")
#                 )
#                 git_branch = (
#                     next(
#                         (
#                             branch
#                             for branch in git_branch_list
#                             if "HEAD" not in branch
#                         ),
#                         "",
#                     )
#                     .replace("origin/", "")
#                     .strip()
#                 )
#             else:
#                 git_branch_from_detached_head_failure = (
#                     "Unable to extract source commit SHA from commit message."
#                 )
#                 raise ValueError(git_branch_from_detached_head_failure)

#         git_short_sha = git("rev-parse", "--short", "HEAD").strip()
#         remote_url = git("config", "--get", "remote.origin.url").strip()
#         repo_name = remote_url.split("/")[-1].rstrip(".git")

#         for string in [repo_name, git_branch, git_short_sha]:
#             if any(char.isupper() for char in string):
#                 logger.warning(
#                     f"String '{string}' contains capitalized characters. Converting to lowercase."
#                 )

#         return repo_name.lower(), git_branch.lower(), git_short_sha.lower()

#     except plumbum.commands.processes.ProcessExecutionError as e:
#         logger.error(f"Error obtaining git information: {e}")
#         raise



"""Tests for `pyrovelocity._velocity_model` module."""

import pytest
import torch
from pyro.distributions import Poisson
from pyro.nn import PyroModule

from pyrovelocity.models._velocity_model import LogNormalModel
from pyrovelocity.models._velocity_model import VelocityModelAuto


def test_load__velocity_model():
    from pyrovelocity.models import _velocity_model

    print(_velocity_model.__file__)


class TestLogNormalModel:
    @pytest.fixture
    def log_normal_model(self):
        return LogNormalModel(
            num_cells=3, num_genes=4, likelihood="Poisson", plate_size=2
        )

    def test_initialization(self, log_normal_model):
        """Test initialization of LogNormalModel"""
        assert isinstance(log_normal_model, PyroModule)
        assert log_normal_model.num_cells == 3
        assert log_normal_model.num_genes == 4
        assert log_normal_model.likelihood == "Poisson"
        assert log_normal_model.plate_size == 2

    def test_create_plates(self, log_normal_model):
        """Test create_plates method"""
        cell_plate, gene_plate = log_normal_model.create_plates()
        assert cell_plate.size == 3
        assert gene_plate.size == 4

    def test_get_likelihood(self, log_normal_model):
        """Test get_likelihood method"""
        ut = torch.rand(3, 4)
        st = torch.rand(3, 4)
        u_read_depth = torch.rand(3, 1)
        s_read_depth = torch.rand(3, 1)
        u_dist, s_dist = log_normal_model.get_likelihood(
            ut, st, u_read_depth=u_read_depth, s_read_depth=s_read_depth
        )
        assert isinstance(u_dist, Poisson)
        assert isinstance(s_dist, Poisson)

    def test_invalid_likelihood(self, log_normal_model):
        """Test get_likelihood method with invalid likelihood"""
        log_normal_model.likelihood = "Invalid"
        ut = torch.rand(3, 4)
        st = torch.rand(3, 4)
        with pytest.raises(NotImplementedError):
            log_normal_model.get_likelihood(ut, st)


class TestVelocityModelAuto:
    @pytest.fixture
    def velocity_model_auto(self):
        return VelocityModelAuto(
            num_cells=3,
            num_genes=4,
            likelihood="Poisson",
            shared_time=True,
            t_scale_on=False,
            plate_size=2,
            latent_factor="none",
            latent_factor_size=30,
            latent_factor_operation="selection",
            include_prior=False,
            num_aux_cells=100,
            only_cell_times=False,
            decoder_on=False,
            add_offset=False,
            correct_library_size=True,
            guide_type="velocity",
            cell_specific_kinetics=None,
            kinetics_num=None,
        )

    def test_initialization(self, velocity_model_auto):
        """Test initialization of VelocityModelAuto"""
        assert isinstance(velocity_model_auto, LogNormalModel)
        assert velocity_model_auto.num_cells == 3
        assert velocity_model_auto.num_genes == 4
        assert velocity_model_auto.likelihood == "Poisson"
        assert velocity_model_auto.shared_time is True
        assert velocity_model_auto.plate_size == 2

    def test_inherited_create_plates(self, velocity_model_auto):
        """Test inherited create_plates method from LogNormalModel"""
        cell_plate, gene_plate = velocity_model_auto.create_plates()
        assert cell_plate.size == 3
        assert gene_plate.size == 4

    def test_forward_method(self, velocity_model_auto):
        """Test the forward method"""
        u_obs = torch.rand(3, 4)
        s_obs = torch.rand(3, 4)
        u_log_library = torch.tensor([[3.7377], [4.0254], [2.7081]])
        s_log_library = torch.tensor([[3.6376], [3.9512], [2.3979]])
        u_log_library_loc = torch.tensor([[3.4904], [3.4904], [3.4904]])
        s_log_library_loc = torch.tensor([[3.3289], [3.3289], [3.3289]])
        u_log_library_scale = torch.tensor([[0.6926], [0.6926], [0.6926]])
        s_log_library_scale = torch.tensor([[0.8214], [0.8214], [0.8214]])
        ind_x = torch.tensor([2, 0, 1])

        u, s = velocity_model_auto.forward(
            u_obs=u_obs,
            s_obs=s_obs,
            u_log_library=u_log_library,
            s_log_library=s_log_library,
            u_log_library_loc=u_log_library_loc,
            s_log_library_loc=s_log_library_loc,
            u_log_library_scale=u_log_library_scale,
            s_log_library_scale=s_log_library_scale,
            ind_x=ind_x,
        )

        assert u.shape == (3, 4)
        assert s.shape == (3, 4)
